"""
Utility functions for logging, staging data results for external
access (S3 pre-signed URL), decrypting data using a shared secret, and
operating on message queues.

This module relies heavily on environment variables to know
which endpoints to use and how to authenticate to them as follows:

Required when receiving an encrypted user access token in the message (Always!):
    SHARED_SECRET_KEY:  The 32-byte shared encryption / decryption key to decrypt
                        the access token in the Harmony operation.

Required when reading from or staging to S3:
    AWS_DEFAULT_REGION: The AWS region in which the S3 client is operating (default: "us-west-2")

Required when staging to S3 and not using the Harmony-provided stagingLocation prefix:
    STAGING_BUCKET: The bucket where staged files should be placed
    STAGING_PATH: The base path under which staged files should be placed

Always provided by newer versions of the Harmony frontend:
    USER_AGENT:     The Harmony user agent string. E.g. harmony/0.0.0 (harmony-sit)

Optional when reading from or staging to S3:
    USE_LOCALSTACK:  'true' if the S3 client should connect to a LocalStack instance instead of
                     Amazon S3 (for testing)
    LOCALSTACK_HOST: The hostname of the Localstack to connect to if `USE_LOCALSTACK`.
    BACKEND_HOST:    The hostname of the Harmony backend. Deprecated / unused by this package.

Optional:
    APP_NAME:             A name for the service that will appear in log entries.
    ENV:                  The application environment. One of: dev, test. Used for local development.
    TEXT_LOGGER:          Whether to log in plaintext or JSON. Default: True (plaintext).
    MAX_DOWNLOAD_RETRIES: Number of times to retry HTTP download calls that fail due to transient errors.
    POST_URL_LENGTH:      Minimum url length that will be submitted via POST request.
"""

from base64 import b64decode
from collections import namedtuple
from functools import lru_cache
import hashlib
import logging
from pathlib import Path, PurePath
from os import environ, path
import sys
import re
from urllib import parse

from nacl.secret import SecretBox

from harmony_service_lib import aws
from harmony_service_lib import http
# The following imports are for backwards-compatibility for services
# which import them from `harmony_service_lib.util`. Though they are not used in
# this module, importing them here allows applications to work without
# modifications.
from harmony_service_lib.exceptions import (HarmonyException, CanceledException, ForbiddenException)  # noqa: F401
from harmony_service_lib.logging import build_logger
from harmony_service_lib.version import get_version


DEFAULT_SHARED_SECRET_KEY = '_THIS_IS_MY_32_CHARS_SECRET_KEY_'


Config = namedtuple(
    'Config', [
        'app_name',
        'use_localstack',
        'backend_host',
        'localstack_host',
        'aws_default_region',
        'staging_path',
        'staging_bucket',
        'env',
        'text_logger',
        'shared_secret_key',
        'user_agent',
        'max_download_retries',
        'post_url_length'
    ])


def _validated_config(config):
    """Validates that the given Config has values for all required
    variables and returns it if so. Raises an Exception if invalid.
    """
    required = [
        'shared_secret_key',
        'staging_path',
        'staging_bucket',
        'max_download_retries'
    ]

    unset = [var.upper() for var in required if getattr(config, var) is None]

    if len(unset) > 0:
        msg = f"Required environment variables are not set: {', '.join(unset)}"
        raise Exception(msg)

    # Warnings
    if config.shared_secret_key == DEFAULT_SHARED_SECRET_KEY:
        logging.warning("The SHARED_SECRET_KEY has not been set. Currently set to its default (unsecure) value.")

    logging.info(config)

    return config


@lru_cache(maxsize=128)
def config(validate=True):
    """
    Returns the Config object with all parameters set to values that were set in the
    process' environment (as environment variables), or to their default values if not
    set.

    Parameters
    ----------
    validate : bool
        Whether to validate the config before returning it. Useful to disable when
        running unit tests.

    Returns
    -------
    harmony_service_lib.util.Config
        The configuration values for this runtime environment.
    """
    def str_envvar(name: str, default: str) -> str:
        value = environ.get(name, default)
        return value.strip('\"') if value is not None else None

    def bool_envvar(name: str, default: bool) -> bool:
        value = environ.get(name)
        return str.lower(value) == 'true' if value is not None else default

    def int_envvar(name: str, default: int) -> int:
        value = environ.get(name)
        return int(value) if value is not None else default

    backend_host = str_envvar('BACKEND_HOST', 'localhost')
    localstack_host = str_envvar('LOCALSTACK_HOST', backend_host)

    config = Config(
        app_name=str_envvar('APP_NAME', sys.argv[0]),
        use_localstack=bool_envvar('USE_LOCALSTACK', False),
        backend_host=backend_host,
        localstack_host=localstack_host,
        aws_default_region=str_envvar('AWS_DEFAULT_REGION', 'us-west-2'),
        staging_path=str_envvar('STAGING_PATH', None),
        staging_bucket=str_envvar('STAGING_BUCKET', None),
        env=str_envvar('ENV', ''),
        text_logger=bool_envvar('TEXT_LOGGER', False),
        shared_secret_key=str_envvar('SHARED_SECRET_KEY', DEFAULT_SHARED_SECRET_KEY),
        user_agent=str_envvar('USER_AGENT', 'harmony (unknown version)'),
        max_download_retries=int_envvar('MAX_DOWNLOAD_RETRIES', 0),
        post_url_length=int_envvar('POST_URL_LENGTH', 2000)
    )

    if validate:
        return _validated_config(config)
    else:
        return config


def _build_full_user_agent(config) -> str:
    """
    Builds a user-agent string that can be passed on to aws or http clients.
    The user agent may consist of a user agent defined by an env variable passed
    by newer versions of Harmony, a user agent for this service lib, and an optional user
    agent that can be provided by users of this lib.

    Parameters
    ----------
    config : harmony_service_lib.util.Config
        The configuration values for this runtime environment.

    Returns
    -------
    string
        A user agent string.
    """
    harmony_user_agent = config.user_agent
    app_name = config.app_name
    lib_user_agent = f'harmony-service-lib/{get_version()}'
    full_user_agent = f'{harmony_user_agent} {lib_user_agent}'
    if app_name is not None:
        full_user_agent += f' ({app_name})'
    return full_user_agent


def _is_file_url(url: str) -> bool:
    return url is not None and url.startswith('file://')


def _url_as_filename(url: str) -> str:
    """Return a version of the url optimized for local development.

    If the url is a `file://` url, it will return the remaining part
    of the url so it can be used as a local file path. For example,
    'file:///logs/example.txt' will be converted to
    '/logs/example.txt'.

    Parameters
    ----------
    url: str The url to check and optaimize.
    Returns
    -------
    str: The url converted to a filename.

    """
    return url.replace('file://', '')


def _filename(directory_path: str, url: str) -> Path:
    """Constructs a filename from the url using the specified directory
    as its path. The constructed filename will be a sha256 hash
    (converted to a hex digest) of the url, and the file's extension
    will be the same as that of the filename in the url.

    Parameters
    ----------
    directory_path : str
        The url to use when constructing the filename and extension.
    Returns
    -------

    """
    return Path(
        directory_path,
        hashlib.sha256(url.encode('utf-8')).hexdigest()
    ).with_suffix(PurePath(parse.urlparse(url).path).suffix)


def download(url, destination_dir, logger=None, access_token=None, data=None, cfg=None):
    """
    Downloads the given URL to the given destination directory, using the basename of the URL
    as the filename in the destination directory.  Supports http://, https:// and s3:// schemes.
    When using the s3:// scheme, will run against us-west-2 unless the "AWS_DEFAULT_REGION"
    environment variable is set.

    When using http:// or https:// schemes, the access_token will be used for authentication.

    Parameters
    ----------
    url : string
        The URL to fetch
    destination_dir : string
        The directory in which to place the downloaded file
    logger : Logger
        A logger to which the function will write, if provided
    access_token :
        The Earthdata Login token of the caller to use for downloads
    data : dict or Tuple[str, str]
        Optional parameter for additional data to
        send to the server when making a HTTP POST request through
        urllib.get.urlopen. These data will be URL encoded to a query string
        containing a series of `key=value` pairs, separated by ampersands. If
        None (the default), urllib.get.urlopen will use the  GET
        method.
    cfg : harmony_service_lib.util.Config
        The configuration values for this runtime environment.

    Returns
    -------
    destination : string
      The filename, including directory, of the downloaded file
    """
    if cfg is None:
        cfg = config()
    if logger is None:
        logger = build_logger(cfg)

    if _is_file_url(url):
        return _url_as_filename(url)

    source = http.localhost_url(url, cfg.localstack_host)

    destination_path = _filename(destination_dir, url)
    if destination_path.exists():
        return str(destination_path)
    destination_path = str(destination_path)

    full_user_agt = _build_full_user_agent(cfg)

    with open(destination_path, 'wb') as destination_file:
        if aws.is_s3(source):
            aws.download(cfg, source, destination_file, full_user_agt)
        elif http.is_http(source):
            http.download(cfg, source, access_token, data, destination_file, full_user_agt)
        else:
            msg = f'Unable to download a url of unknown type: {url}'
            logger.error(msg)
            raise Exception(msg)

    return destination_path


def stage(local_filename, remote_filename, mime, logger=None, location=None, cfg=None):
    """
    Stages the given local filename, including directory path, to an S3 location with the given
    filename and mime-type

    Requires the following environment variables:
        AWS_DEFAULT_REGION: The AWS region in which the S3 client is operating

    Parameters
    ----------
    local_filename : string
        A path and filename to the local file that should be staged
    remote_filename : string
        The basename to give to the remote file
    mime : string
        The mime type to apply to the staged file for use when it is served, e.g. "application/x-netcdf4"
    location : string
        The S3 prefix URL under which to place the output file.  If not provided, STAGING_BUCKET and
        STAGING_PATH must be set in the environment
    logger : logging
        The logger to use
    cfg : harmony_service_lib.util.Config
        The configuration values for this runtime environment.

    Returns
    -------
    url : string
        An s3:// URL to the staged file
    """
    # The implementation of this function has been moved to the
    # harmony_service_lib.aws module.
    if cfg is None:
        cfg = config()
    if logger is None:
        logger = build_logger(cfg)

    return aws.stage(cfg, local_filename, remote_filename, mime, logger, location)


def create_decrypter(key=b'_THIS_IS_MY_32_CHARS_SECRET_KEY_'):
    """Creates a function that will decrypt cyphertext using a shared secret
    (symmetric) 32-byte key.

    The returned decrypter function has type signature: str -> str.
    """
    box = SecretBox(key)

    def decrypter(encrypted_msg_str):
        """Decrypt encrypted text using the shared secret (symmetric) key
        in the function's closure."""

        parts = encrypted_msg_str.split(':')
        nonce = b64decode(parts[0])
        ciphertext = b64decode(parts[1])

        return box.decrypt(ciphertext, nonce).decode('utf-8')

    return decrypter


def nop_decrypter(ciphertext):
    """An identity decrypter function. A NOP: it returns the ciphertext
    as-is.
    """
    return ciphertext


def generate_output_filename(
    filename: str,
    ext: str = None,
    variable_subset: list[str] = None,
    is_regridded: bool = False,
    is_subsetted: bool = False,
    is_reformatted: bool = False
) -> str:
    """
    Return an output filename for the given granules according to our naming conventions:
    {original filename without suffix}(_{single var})?(_regridded)?(_subsetted)?(_reformatted)?.<ext>

    Parameters
    ----------
        granule : message.Granule
            The source granule for the output file
        ext: string, optional
            The destination file extension (default: original extension)
        variable_subset : string[], optional
            When variable subsetting, a list of all variables that have been subset
        is_regridded : bool, optional
            True if a regridding operation has been performed (default: False)
        is_subsetted : bool, optional
            True if a subsetting operation has been performed (default: False)
        is_reformatted : bool, optional
            True if a reformatting operation has been performed (default: False)

    Returns
    -------
        string
            The output filename
    """
    url = filename
    # Get everything between the last non-trailing '/' before the query and the first '?'
    # Do this instead of using a URL parser, because our URLs are not complex in practice and
    # it is useful to allow relative file paths to work for local testing.
    original_filename = url.split('?')[0].rstrip('/').split('/')[-1]
    decoded_original_filename = parse.unquote(original_filename)
    (original_basename, original_ext) = path.splitext(decoded_original_filename)
    if ext is None:
        ext = original_ext

    if not ext.startswith('.'):
        ext = '.' + ext

    suffixes = []
    if variable_subset and len(variable_subset) == 1:
        var = variable_subset[0]
        if hasattr(var, 'name'):
            var = var.name
        suffixes.append('_' + var)
    if is_regridded:
        suffixes.append('_regridded')
    if is_subsetted:
        suffixes.append('_subsetted')
    if is_reformatted:
        suffixes.append('_reformatted')

    suffixes.append(ext)

    result = original_basename
    # Iterate suffixes in reverse, removing them from the result if they're at the end of the string
    # This supports the case of chaining where one service regrids and another subsets but we don't
    # want names to get mangled
    for suffix in suffixes[::-1]:
        if result.endswith(suffix):
            result = result[:-len(suffix)]

    result += "".join(suffixes)

    # replace special chars that may have been encoded or present in variable_subset
    result = re.sub('\\/|:', '_', result)

    # runs of underscores are replaced with single underscore
    result = re.sub(r'_{2,}', '_', result)

    # leading or trailing underscores are removed
    result = re.sub(r'^_+|_+$', '', result)

    # underscores before or after periods are removed
    result = re.sub(r'_{0,}\._{0,}', '.', result)

    return result


def bbox_to_geometry(bbox):
    '''
    Creates a GeoJSON geometry given a GeoJSON BBox, accounting for antimeridian

    Parameters
    ----------
    bbox : float[4]
        the bounding box to create a geometry from

    Returns
    -------
    dict
        a GeoJSON Polygon or MultiPolygon representation of the input bbox
    '''
    if not bbox:
        return None
    west, south, east, north = bbox[0:4]
    if west > east:
        return {
            'type': 'MultiPolygon',
            'coordinates': [
                [[
                    [-180, south],
                    [-180, north],
                    [east, north],
                    [east, south],
                    [-180, south]
                ]],
                [[
                    [west, south],
                    [west, north],
                    [180, north],
                    [180, south],
                    [west, south]
                ]]
            ]
        }
    return {
        'type': 'Polygon',
        'coordinates': [[
            [west, south],
            [west, north],
            [east, north],
            [east, south],
            [west, south]
        ]],
    }



class HarmonyException(Exception):
    """Base class for Harmony exceptions.

    Attributes
    ----------
    message : string
        Explanation of the error
    category : string
        Classification of the type of harmony error
    level : string
        The level of the error, can be 'Error' or 'Warning'.
    """

    def __init__(self, message, category='Service', level='Error'):
        self.message = message
        self.category = category
        self.level = level


class CanceledException(HarmonyException):
    """Class for throwing an exception indicating a Harmony request has been canceled"""

    def __init__(self, message=None):
        super().__init__(message, 'Canceled')


class ForbiddenException(HarmonyException):
    """Class for throwing an exception indicating download failed due to not being able to access the data"""

    def __init__(self, message=None):
        super().__init__(message, 'Forbidden')


class ServerException(HarmonyException):
    """Class for throwing an exception indicating the download failed due to a generic 500 internal server error """

    def __init__(self, message=None):
        super().__init__(message, 'Server')


class NoRetryException(HarmonyException):
    """Class for throwing an exception indicating the error is a fatal error that is specific to a request
    and should not be retried by Harmony."""

    def __init__(self, message=None):
        super().__init__(message, 'NoRetry')


class NoDataException(HarmonyException):
    """Class for throwing an exception indicating service found no data to process """

    def __init__(self, message='No data found to process'):
        super().__init__(message, 'NoData', 'Warning')



from unittest import TestCase

from harmony_service_lib.message import Message

from harmony_service_lib.message_utility import (
    _has_all_attributes,
    _has_consistent_dimension,
    has_crs,
    has_dimensions,
    has_self_consistent_grid,
    has_scale_extents,
    has_scale_sizes,
    has_valid_scale_extents,
    rgetattr,
)


class TestMessageUtility(TestCase):
    """Test Harmony Message utilities."""


    def test_self_has_self_consistent_grid(self):
        """ Ensure that the function correctly determines if the supplied
            Harmony message defines a valid target grid. This should either:

            * Specify scaleExtent and 1 of:
              * Height and Width
              * Scale sizes (in the x and y horizontal spatial dimensions)
            * Specify all three of the above, but the values must be consistent
              with one another.

        """
        valid_scale_extents = {'x': {'min': -180, 'max': 180},
                               'y': {'min': -90, 'max': 90}}

        valid_scale_sizes = {'x': 0.5, 'y': 1.0}
        valid_height = 180
        valid_width = 720

        with self.subTest('format = None returns False'):
            test_message = Message({})
            self.assertFalse(has_self_consistent_grid(test_message))

        with self.subTest('No grid params, allow_incomplete_grid = True returns True'):
            test_message = Message({})
            self.assertTrue(has_self_consistent_grid(test_message, True))

        with self.subTest('All grid parameters = None returns False'):
            test_message = Message({'format': {}})
            self.assertFalse(has_self_consistent_grid(test_message))

        with self.subTest('Only scaleExtent returns False'):
            test_message = Message({
                'format': {'scaleExtents': valid_scale_extents}
            })
            self.assertFalse(has_self_consistent_grid(test_message))

        with self.subTest('Only scaleSize returns False'):
            test_message = Message({
                'format': {'scaleSize': valid_scale_sizes}
            })
            self.assertFalse(has_self_consistent_grid(test_message))

        with self.subTest('Only dimensions (height and width) returns False'):
            test_message = Message({
                'format': {'height': valid_height, 'width': valid_width}
            })
            self.assertFalse(has_self_consistent_grid(test_message))

        with self.subTest('Dimensions and scaleSize returns False'):
            # Need the extents to know where to put the pixels!
            test_message = Message({
                'format': {'height': valid_height,
                           'scaleSize': valid_scale_sizes,
                           'width': valid_width}
            })
            self.assertFalse(has_self_consistent_grid(test_message))

        with self.subTest('Dimensions and scaleExtent returns True'):
            test_message = Message({
                'format': {'height': valid_height,
                           'scaleExtent': valid_scale_extents,
                           'width': valid_width}
            })
            self.assertTrue(has_self_consistent_grid(test_message))

        with self.subTest('scaleExtent and scaleSize returns True'):
            test_message = Message({
                'format': {'scaleExtent': valid_scale_extents,
                           'scaleSize': valid_scale_sizes}
            })
            self.assertTrue(has_self_consistent_grid(test_message))

        with self.subTest('All three, and consistent returns True'):
            test_message = Message({
                'format': {'height': valid_height,
                           'scaleExtent': valid_scale_extents,
                           'scaleSize': valid_scale_sizes,
                           'width': valid_width}
            })
            self.assertTrue(has_self_consistent_grid(test_message))

        with self.subTest('All three, but inconsistent returns False'):
            test_message = Message({
                'format': {'height': valid_height + 150,
                           'scaleExtent': valid_scale_extents,
                           'scaleSize': valid_scale_sizes,
                           'width': valid_width - 150}
            })
            self.assertFalse(has_self_consistent_grid(test_message))

        with self.subTest('Inconsistent grid, allow_incomplete_grid=True, returns False'):
            test_message = Message({
                'format': {'height': valid_height + 150,
                           'scaleExtent': valid_scale_extents,
                           'scaleSize': valid_scale_sizes,
                           'width': valid_width - 150}
            })
            self.assertFalse(has_self_consistent_grid(test_message, True))

    def test_self_has_self_consistent_grid_missing_height_or_width(self):
        """ Ensure that the function correctly determines if the supplied
            Harmony message defines a valid target grid. These test cases check
            when only one of height or width are specified.

            If there is sufficient other grid information, and the one listed
            dimension with all three pieces of information is self consistent,
            then the message passes validation.

            If there are grids with insufficient information, but
            `allow_incomplete_grid=True`, then the validation should pass for
            under-defined grids.

        """
        valid_scale_extents = {'x': {'min': -180, 'max': 180},
                               'y': {'min': -90, 'max': 90}}

        valid_scale_sizes = {'x': 0.5, 'y': 1.0}
        valid_height = 180
        valid_width = 720

        with self.subTest('Only height and scaleExtent returns False'):
            test_message = Message({
                'format': {'height': valid_height,
                           'scaleExtent': valid_scale_extents}
            })
            self.assertFalse(has_self_consistent_grid(test_message))

        with self.subTest('Only width and scaleExtent returns False'):
            test_message = Message({
                'format': {'scaleExtent': valid_scale_extents,
                           'width': valid_width}
            })
            self.assertFalse(has_self_consistent_grid(test_message))

        with self.subTest('Only height and scaleSize returns False'):
            test_message = Message({
                'format': {'height': valid_height,
                           'scaleSize': valid_scale_sizes}
            })
            self.assertFalse(has_self_consistent_grid(test_message))

        with self.subTest('Only width and scaleSize returns False'):
            test_message = Message({
                'format': {'scaleSize': valid_scale_sizes,
                           'width': valid_width}
            })
            self.assertFalse(has_self_consistent_grid(test_message))

        with self.subTest('Width, scaleSize, scaleExtent inconsistent, False'):
            test_message = Message({
                'format': {'scaleExtent': valid_scale_extents,
                           'scaleSize': valid_scale_sizes,
                           'width': valid_width - 150}
            })
            self.assertFalse(has_self_consistent_grid(test_message))

        with self.subTest('Height, scaleSize, scaleExtent inconsistent, False'):
            test_message = Message({
                'format': {'height': valid_height + 150,
                           'scaleExtent': valid_scale_extents,
                           'scaleSize': valid_scale_sizes}
            })
            self.assertFalse(has_self_consistent_grid(test_message))

        with self.subTest('Width, scaleSize, scaleExtent consistent, True'):
            test_message = Message({
                'format': {'scaleExtent': valid_scale_extents,
                           'scaleSize': valid_scale_sizes,
                           'width': valid_width}
            })
            self.assertTrue(has_self_consistent_grid(test_message))

        with self.subTest('Height, scaleSize, scaleExtent consistent, True'):
            test_message = Message({
                'format': {'height': valid_height,
                           'scaleExtent': valid_scale_extents,
                           'scaleSize': valid_scale_sizes}
            })
            self.assertTrue(has_self_consistent_grid(test_message))

        with self.subTest('Only height and scaleExtent, uses allow_incomplete_grid=True'):
            test_message = Message({
                'format': {'height': valid_height,
                           'scaleExtent': valid_scale_extents}
            })
            self.assertTrue(
                has_self_consistent_grid(test_message, allow_incomplete_grid=True)
            )

        with self.subTest('Only width and scaleExtent, uses allow_incomplete_grid=True'):
            test_message = Message({
                'format': {'scaleExtent': valid_scale_extents,
                           'width': valid_width}
            })
            self.assertTrue(
                has_self_consistent_grid(test_message, allow_incomplete_grid=True)
            )

        with self.subTest('Only height and scaleSize, uses allow_incomplete_grid=True'):
            test_message = Message({
                'format': {'height': valid_height,
                           'scaleSize': valid_scale_sizes}
            })
            self.assertTrue(
                has_self_consistent_grid(test_message, allow_incomplete_grid=True)
            )

        with self.subTest('Only width and scaleSize, usses allow_incomplete_grid=True'):
            test_message = Message({
                'format': {'scaleSize': valid_scale_sizes,
                           'width': valid_width}
            })
            self.assertTrue(
                has_self_consistent_grid(test_message, allow_incomplete_grid=True)
            )

    def test_message_has_crs(self):
        message = Message({'format': {'crs': 'EPSG:4326'}})
        self.assertTrue(has_crs(message))

    def test_message_has_garbage_crs(self):
        message = Message({'format': {'crs': 'garbage'}})
        self.assertTrue(has_crs(message))

    def test_message_has_no_crs(self):
        message = Message({})
        self.assertFalse(has_crs(message))

    def test_has_consistent_dimension(self):
        """Ensure, given a scale size (resolution), scale extent (range) and
        dimension size, the function can correctly determine if all three
        values are consistent with one another.

        """
        valid_scale_extents = {
            'x': {'min': -180, 'max': 180},
            'y': {'min': -90, 'max': 90},
        }

        valid_scale_sizes = {'x': 0.5, 'y': 1.0}
        valid_height = 180
        valid_width = 720

        with self.subTest('Consistent x dimension returns True'):
            test_message = Message(
                {
                    'format': {
                        'scaleExtent': valid_scale_extents,
                        'scaleSize': valid_scale_sizes,
                        'width': valid_width,
                    }
                }
            )
            self.assertTrue(_has_consistent_dimension(test_message, 'x'))

        with self.subTest('Consistent y dimension returns True'):
            test_message = Message(
                {
                    'format': {
                        'scaleExtent': valid_scale_extents,
                        'scaleSize': valid_scale_sizes,
                        'height': valid_height,
                    }
                }
            )
            self.assertTrue(_has_consistent_dimension(test_message, 'y'))

        with self.subTest('Inconsistent x dimension returns False'):
            test_message = Message(
                {
                    'format': {
                        'scaleExtent': valid_scale_extents,
                        'scaleSize': valid_scale_sizes,
                        'width': valid_width + 100,
                    }
                }
            )
            self.assertFalse(_has_consistent_dimension(test_message, 'x'))

        with self.subTest('Inconsistent y dimension returns False'):
            test_message = Message(
                {
                    'format': {
                        'scaleExtent': valid_scale_extents,
                        'scaleSize': valid_scale_sizes,
                        'height': valid_height + 100,
                    }
                }
            )
            self.assertFalse(_has_consistent_dimension(test_message, 'y'))

    def test_has_all_attributes(self):
        """Ensure that the function returns the correct boolean value
        indicating if all requested attributes are present in the supplied
        object, and have non-None values.

        """
        test_message = Message({'format': {'scaleSize': {'x': 0.5, 'y': 0.5}}})

        with self.subTest('All attributes present returns True'):
            self.assertTrue(
                _has_all_attributes(
                    test_message, ['format.scaleSize.x', 'format.scaleSize.y']
                )
            )

        with self.subTest('Some attributes present returns False'):
            self.assertFalse(
                _has_all_attributes(
                    test_message, ['format.scaleSize.x', 'format.height']
                )
            )

        with self.subTest('No attributes present returns False'):
            self.assertFalse(
                _has_all_attributes(test_message, ['format.height', 'format.width'])
            )

    def test_has_scale_sizes(self):
        """Ensure the function correctly identifies whether the supplied
        Harmony message contains both an x and y scale size.

        """
        with self.subTest('Scale sizes present returns True'):
            test_message = Message({'format': {'scaleSize': {'x': 0.5, 'y': 0.5}}})
            self.assertTrue(has_scale_sizes(test_message))

        with self.subTest('scaleSize.x = None returns False'):
            test_message = Message({'format': {'scaleSize': {'y': 0.5}}})
            self.assertFalse(has_scale_sizes(test_message))

        with self.subTest('scaleSize.y = None returns False'):
            test_message = Message({'format': {'scaleSize': {'x': 0.5}}})
            self.assertFalse(has_scale_sizes(test_message))

        with self.subTest('Both scaleSizes = None returns False'):
            test_message = Message({'format': {'scaleSize': {}}})
            self.assertFalse(has_scale_sizes(test_message))

        with self.subTest('format = None returns False'):
            test_message = Message({})
            self.assertFalse(has_scale_sizes(test_message))

    def test_has_valid_scale_extents(self):
        with self.subTest('ScaleExtent present and valid returns True'):
            test_message = Message(
                {
                    'format': {
                        'scaleExtent': {
                            'x': {'min': -180, 'max': 180},
                            'y': {'min': -90, 'max': 90},
                        }
                    }
                }
            )
            self.assertTrue(has_valid_scale_extents(test_message))

        with self.subTest('ScaleExtent missing returns True'):
            test_message = Message({'format': {}})
            self.assertTrue(has_valid_scale_extents(test_message))

        with self.subTest('ScaleExtent present and invalid returns False'):
            test_message = Message(
                {
                    'format': {
                        'scaleExtent': {
                            'x': {'min': 180, 'max': -180},
                            'y': {'min': -90, 'max': 90},
                        }
                    }
                }
            )
            self.assertFalse(has_valid_scale_extents(test_message))

    def test_has_scale_extents(self):
        """Ensure the function correctly identifies whether the supplied
        Harmony message contains all required elements in the
        `format.scaleExtent` attribute. This includes minima and maxima for
        both the x and y horizontal spatial dimensions of the target grid.

        """
        x_extents = {'min': -180, 'max': 180}
        y_extents = {'min': -90, 'max': 90}

        with self.subTest('Scale extents present returns True'):
            test_message = Message(
                {'format': {'scaleExtent': {'x': x_extents, 'y': y_extents}}}
            )
            self.assertTrue(has_scale_extents(test_message))

        with self.subTest('scaleExtent.x.min = None returns False'):
            test_message = Message(
                {'format': {'scaleExtent': {'x': {'max': 180}, 'y': y_extents}}}
            )
            self.assertFalse(has_scale_extents(test_message))

        with self.subTest('scaleExtent.x.max = None returns False'):
            test_message = Message(
                {'format': {'scaleExtent': {'x': {'min': -180}, 'y': y_extents}}}
            )
            self.assertFalse(has_scale_extents(test_message))

        with self.subTest('scaleExtent.x min and max = None returns False'):
            test_message = Message(
                {'format': {'scaleExtent': {'x': {}, 'y': y_extents}}}
            )
            self.assertFalse(has_scale_extents(test_message))

        with self.subTest('scaleExtent.y.min = None returns False'):
            test_message = Message(
                {'format': {'scaleExtent': {'x': x_extents, 'y': {'max': 90}}}}
            )
            self.assertFalse(has_scale_extents(test_message))

        with self.subTest('scaleExtent.y.max = None returns False'):
            test_message = Message(
                {'format': {'scaleExtent': {'x': x_extents, 'y': {'min': -90}}}}
            )
            self.assertFalse(has_scale_extents(test_message))

        with self.subTest('scaleExtent.y min and max = None returns False'):
            test_message = Message(
                {'format': {'scaleExtent': {'x': x_extents, 'y': {}}}}
            )
            self.assertFalse(has_scale_extents(test_message))

        with self.subTest('All scaleExtent values = None returns False'):
            test_message = Message({'format': {'scaleExtent': {'x': {}, 'y': {}}}})
            self.assertFalse(has_scale_extents(test_message))

        with self.subTest('scaleExtent.x and scaleExtent.y = None returns False'):
            test_message = Message({'format': {'scaleExtent': {}}})
            self.assertFalse(has_scale_extents(test_message))

        with self.subTest('format = None returns False'):
            test_message = Message({})
            self.assertFalse(has_scale_extents(test_message))

    def test_has_dimensions(self):
        """Ensure the function correctly validates whether the supplied
        Harmony message contains both a height an width for the target
        grid.

        """
        with self.subTest('height and width present returns True'):
            test_message = Message({'format': {'height': 100, 'width': 50}})
            self.assertTrue(has_dimensions(test_message))

        with self.subTest('height = None returns False'):
            test_message = Message({'format': {'width': 50}})
            self.assertFalse(has_dimensions(test_message))

        with self.subTest('width = None returns False'):
            test_message = Message({'format': {'height': 100}})
            self.assertFalse(has_dimensions(test_message))

        with self.subTest('height = None and width = None returns False'):
            test_message = Message({'format': {}})
            self.assertFalse(has_dimensions(test_message))

        with self.subTest('format = None returns False'):
            test_message = Message({})
            self.assertFalse(has_dimensions(test_message))

    def test_rgetattr(self):
        """Ensure that the recursive function can retrieve nested attributes
        and uses the default argument when required.

        """

        class Child:
            def __init__(self, name):
                self.name = name

        class Parent:
            def __init__(self, name, child_name):
                self.name = name
                self.child = Child(child_name)

        test_parent = Parent('parent_name', 'child_name')

        with self.subTest('Parent level attribute'):
            self.assertEqual(rgetattr(test_parent, 'name'), 'parent_name')

        with self.subTest('Nested attribute'):
            self.assertEqual(rgetattr(test_parent, 'child.name'), 'child_name')

        with self.subTest('Missing parent with default'):
            self.assertEqual(rgetattr(test_parent, 'absent', 'default'), 'default')

        with self.subTest('Missing child attribute with default'):
            self.assertEqual(
                rgetattr(test_parent, 'child.absent', 'default'), 'default'
            )

        with self.subTest('Child requested, parent missing, default'):
            self.assertEqual(
                rgetattr(test_parent, 'none.something', 'default'), 'default'
            )

        with self.subTest('Missing parent, with no default'):
            self.assertIsNone(rgetattr(test_parent, 'absent'))

        with self.subTest('Missing child, with no default'):
            self.assertIsNone(rgetattr(test_parent, 'child.absent'))

        with self.subTest('Child requested, parent missing, no default'):
            self.assertIsNone(rgetattr(test_parent, 'absent.something'))



import os
from tempfile import mkdtemp
from datetime import datetime
import shutil
import unittest
import json

from pystac import Catalog, CatalogType, Item

from harmony_service_lib import cli, BaseHarmonyAdapter
from harmony_service_lib.exceptions import ForbiddenException, NoDataException, NoRetryException
from tests.util import cli_parser, config_fixture


class MockAdapter(BaseHarmonyAdapter):
    message = None
    """
    Dummy class to mock adapter calls, performing a no-op service
    """

    def invoke(self):
        MockAdapter.message = self.message
        return (self.message, self.catalog)


class MockMultiCatalogOutputAdapter(BaseHarmonyAdapter):
    message = None
    """
    Dummy class to mock adapter calls, performing a no-op service
    that returns multiple STAC catologs instead of one
    """

    def invoke(self):
        MockAdapter.message = self.message
        catalogs = [
            Catalog('a', ''), Catalog('b', ''), Catalog('c', '')]
        for cat in catalogs:
            items = [
                Item(f'item-1-from-catalog-{cat.id}', None, [0, 0, 1, 1],
                     datetime.strptime('09/19/22 13:55:26', '%m/%d/%y %H:%M:%S'), {}),
                Item(f'item-2-from-catalog-{cat.id}', None, [0, 0, 1, 2],
                     datetime.strptime('09/19/22 13:55:26', '%m/%d/%y %H:%M:%S'), {})
            ]
            cat.add_items(items)
        return (self.message, catalogs)


class TestCliInvokeAction(unittest.TestCase):
    def setUp(self):
        self.workdir = mkdtemp()
        self.inputdir = mkdtemp()
        self.config = config_fixture()
        print(self.config)

    def tearDown(self):
        MockAdapter.messages = []
        shutil.rmtree(self.workdir)

    def test_when_a_service_completes_it_writes_a_output_catalog_to_the_output_dir(self):
        with cli_parser(
                '--harmony-action', 'invoke',
                '--harmony-input', '{"test": "input"}',
                '--harmony-sources', 'example/source/catalog.json',
                '--harmony-metadata-dir', self.workdir) as parser:
            args = parser.parse_args()
            cli.run_cli(parser, args, MockAdapter, cfg=self.config)
            output = Catalog.from_file(os.path.join(self.workdir, 'catalog.json'))
            self.assertTrue(output.validate)

    def test_when_a_service_completes_it_writes_the_output_message_to_the_output_dir(self):
        with cli_parser(
                '--harmony-action', 'invoke',
                '--harmony-input', '{"test": "input"}',
                '--harmony-sources', 'example/source/catalog.json',
                '--harmony-metadata-dir', self.workdir) as parser:
            args = parser.parse_args()
            cli.run_cli(parser, args, MockAdapter, cfg=self.config)
            with open(os.path.join(self.workdir, 'message.json')) as file:
                self.assertEqual(file.read(), '{"test": "input"}')

    def test_when_the_cli_has_a_staging_location_it_overwites_the_message_staging_location(self):
        with cli_parser(
                '--harmony-action', 'invoke',
                '--harmony-input', '{"test": "input"}',
                '--harmony-sources', 'example/source/catalog.json',
                '--harmony-metadata-dir', self.workdir,
                '--harmony-data-location', 's3://fake-location/') as parser:
            args = parser.parse_args()
            cli.run_cli(parser, args, MockAdapter, cfg=self.config)
            self.assertEqual(MockAdapter.message.stagingLocation, 's3://fake-location/')
            # Does not output the altered staging location
            with open(os.path.join(self.workdir, 'message.json')) as file:
                self.assertEqual(file.read(), '{"test": "input"}')

    def test_when_the_backend_service_throws_a_known_error_it_writes_the_error_to_the_output_dir(self):
        with cli_parser(
                '--harmony-action', 'invoke',
                '--harmony-input', '{"test": "input"}',
                '--harmony-sources', 'example/source/catalog.json',
                '--harmony-metadata-dir', self.workdir) as parser:

            class MockImpl(MockAdapter):
                def invoke(self):
                    raise ForbiddenException('Something bad happened')

            args = parser.parse_args()
            with self.assertRaises(Exception) as context:
                cli.run_cli(parser, args, MockImpl, cfg=self.config)

            self.assertTrue('Something bad happened' in str(context.exception))
            with open(os.path.join(self.workdir, 'error.json')) as file:
                self.assertEqual(
                    file.read(),
                    '{"error": "Something bad happened", "category": "Forbidden", "level": "Error"}')

    def test_when_the_backend_service_throws_an_unknown_error_it_writes_a_generic_error_to_the_output_dir(self):
        with cli_parser(
                '--harmony-action', 'invoke',
                '--harmony-input', '{"test": "input"}',
                '--harmony-sources', 'example/source/catalog.json',
                '--harmony-metadata-dir', self.workdir) as parser:

            class MockImpl(MockAdapter):
                def invoke(self):
                    raise Exception('Something bad happened')

            args = parser.parse_args()
            with self.assertRaises(Exception) as context:
                cli.run_cli(parser, args, MockImpl, cfg=self.config)

            self.assertTrue('Something bad happened' in str(context.exception))
            with open(os.path.join(self.workdir, 'error.json')) as file:
                self.assertEqual(
                    file.read(),
                    '{"error": "Service request failed with an unknown error", "category": "Unknown", "level": "Error"}')

    def test_when_the_backend_service_throws_a_no_retry_error_it_writes_the_error_to_the_output_dir(self):
        with cli_parser(
                '--harmony-action', 'invoke',
                '--harmony-input', '{"test": "input"}',
                '--harmony-sources', 'example/source/catalog.json',
                '--harmony-metadata-dir', self.workdir) as parser:

            class MockImpl(MockAdapter):
                def invoke(self):
                    raise NoRetryException('Not a retriable error, no need to retry.')

            args = parser.parse_args()
            with self.assertRaises(NoRetryException) as context:
                cli.run_cli(parser, args, MockImpl, cfg=self.config)

            self.assertTrue('Not a retriable error, no need to retry.' in str(context.exception))
            with open(os.path.join(self.workdir, 'error.json')) as file:
                self.assertEqual(
                    file.read(),
                    '{"error": "Not a retriable error, no need to retry.", "category": "NoRetry", "level": "Error"}')

    def test_when_the_backend_service_throws_a_known_warning_it_writes_the_warning_to_the_output_dir(self):
        with cli_parser(
                '--harmony-action', 'invoke',
                '--harmony-input', '{"test": "input"}',
                '--harmony-sources', 'example/source/catalog.json',
                '--harmony-metadata-dir', self.workdir) as parser:

            class MockImpl(MockAdapter):
                def invoke(self):
                    raise NoDataException('There is no data found')

            args = parser.parse_args()
            with self.assertRaises(Exception) as context:
                cli.run_cli(parser, args, MockImpl, cfg=self.config)

            self.assertTrue('There is no data found' in str(context.exception))
            with open(os.path.join(self.workdir, 'error.json')) as file:
                self.assertEqual(
                    file.read(),
                    '{"error": "There is no data found", "category": "NoData", "level": "Warning"}')

    def test_when_multi_catalog_output_it_saves_with_particular_layout(self):
        with cli_parser(
                '--harmony-action', 'invoke',
                '--harmony-input', '{"test": "input"}',
                '--harmony-sources', 'example/source/catalog.json',
                '--harmony-metadata-dir', self.workdir) as parser:
            args = parser.parse_args()
            cli.run_cli(parser, args, MockMultiCatalogOutputAdapter, cfg=self.config)
            for idx in range(3):
                cat = Catalog.from_file(os.path.join(self.workdir, f'catalog{idx}.json'))
                cat_root = cat.get_single_link('root')
                self.assertEqual(cat_root.get_href(), f'./catalog{idx}.json')
                item_hrefs = [l.get_href() for l in cat.get_links('item')]
                self.assertTrue(f'./item-1-from-catalog-{cat.id}/item-1-from-catalog-{cat.id}.json' in item_hrefs)
                self.assertTrue(f'./item-2-from-catalog-{cat.id}/item-2-from-catalog-{cat.id}.json' in item_hrefs)
                item = Item.from_file(os.path.join(
                    self.workdir, f'./item-1-from-catalog-{cat.id}/item-1-from-catalog-{cat.id}.json'))
                item_root_href = item.get_single_link('root').get_href()
                item_parent_href = item.get_single_link('parent').get_href()
                self.assertTrue(item_parent_href == item_root_href)
                self.assertEqual(item_root_href, f'../catalog{idx}.json')
                self.assertEqual(item_parent_href, f'../catalog{idx}.json')
            with open(os.path.join(self.workdir, 'batch-count.txt')) as file:
                self.assertEqual(file.read(), '3')
            with open(os.path.join(self.workdir, 'batch-catalogs.json')) as file:
                self.assertEqual(json.loads(file.read()),
                                 ["catalog0.json",
                                  "catalog1.json",
                                  "catalog2.json"])


if __name__ == '__main__':
    unittest.main()



import harmony_service_lib


def get_version():
    """
    Get the version of the currently running service lib.

    Returns
    -------
    string
        A string representing the current version.
    """
    return harmony_service_lib.__version__



import datetime as dt
from functools import lru_cache
import logging
import sys
import copy

from pythonjsonlogger import jsonlogger
from harmony_service_lib import message


class HarmonyJsonFormatter(jsonlogger.JsonFormatter):
    """A JSON log entry formatter."""
    def add_fields(self, log_record, record, message_dict):
        super(HarmonyJsonFormatter, self).add_fields(
            log_record, record, message_dict)
        if not log_record.get('timestamp'):
            now = dt.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
            log_record['timestamp'] = now
        if log_record.get('level'):
            log_record['level'] = log_record['level'].upper()
        else:
            log_record['level'] = record.levelname
        if not log_record.get('application'):
            log_record['application'] = self.app_name


class RedactorFormatter(object):
    """Redacts sensitive information from logs."""
    def __init__(self, original_formatter):
        self.original_formatter = original_formatter

    def format(self, record):
        # need to check the log record's msg and args for sensitive values
        # https://docs.python.org/3/library/logging.html#logrecord-attributes
        msg_clone = None
        args_clone = None
        if isinstance(record.msg, message.Message):
            msg_clone = copy.deepcopy(record.msg)
            msg_clone.accessToken = '<redacted>'
        if isinstance(record.args, dict):
            for k in record.args.keys():
                if isinstance(record.args[k], message.Message):
                    if not args_clone:
                        args_clone = copy.deepcopy(record.args)
                    args_clone[k].accessToken = '<redacted>'
        else:  # args is a tuple
            for index, arg in enumerate(record.args):
                if isinstance(arg, message.Message):
                    if not args_clone:
                        args_clone = copy.deepcopy(record.args)
                    args_clone[index].accessToken = '<redacted>'
        if msg_clone:
            record.msg = msg_clone
        if args_clone:
            record.args = args_clone
        formatted_message = self.original_formatter.format(record)
        return formatted_message

    def __getattr__(self, attr):
        return getattr(self.original_formatter, attr)


@lru_cache(maxsize=128)
def build_logger(config, name='harmony-service', stream=None):
    """
    Builds a logger with appropriate defaults for Harmony
    Parameters
    ----------
    config : harmony_service_lib.util.Config
        The configuration values for this runtime environment.
    name : string
        The name of the logger
    stream :
        The stream to write to (optional)

    Returns
    -------
    logger : Logging
        A logger for service output
    """
    logger = logging.getLogger(name)
    syslog = logging.StreamHandler(stream)
    if config.text_logger:
        formatter = logging.Formatter("%(asctime)s [%(levelname)s] [%(name)s.%(funcName)s:%(lineno)d] %(message)s")
    else:
        formatter = HarmonyJsonFormatter()
        formatter.app_name = config.app_name
    syslog.setFormatter(RedactorFormatter(formatter))
    logger.addHandler(syslog)
    logger.setLevel(logging.INFO)
    logger.propagate = False
    return logger


def setup_stdout_log_formatting(config):
    """
    Updates sys.stdout and sys.stderr to pass messages through the Harmony log formatter.
    """
    # See https://stackoverflow.com/questions/11124093/redirect-python-print-output-to-logger/11124247
    class StreamToLogger(object):
        def __init__(self, logger, log_level=logging.INFO):
            self.logger = logger
            self.log_level = log_level
            self.linebuf = ''

        def write(self, buf):
            temp_linebuf = self.linebuf + buf
            self.linebuf = ''
            for line in temp_linebuf.splitlines(True):
                if line[-1] == '\n':
                    self.logger.log(self.log_level, line.rstrip())
                else:
                    self.linebuf += line

        def flush(self):
            if self.linebuf != '':
                self.logger.log(self.log_level, self.linebuf.rstrip())
            self.linebuf = ''
    sys.stdout = StreamToLogger(build_logger(config), logging.INFO)
    sys.stderr = StreamToLogger(build_logger(config), logging.ERROR)



from base64 import b64encode
import unittest

from nacl.secret import SecretBox
from nacl.utils import random

from harmony_service_lib import util


class TestDecrypter(unittest.TestCase):
    def test_when_using_nop_decrypter_the_plaintext_is_the_same_as_cyphertext(self):
        decrypter = util.nop_decrypter
        cyphertext = 'This is a terribly encrypted message.'
        expected = cyphertext

        actual = decrypter(cyphertext)

        self.assertEqual(actual, expected)

    def test_when_encrypting_with_a_key_the_decrypter_works_when_using_the_shared_key(self):
        nonce = random(SecretBox.NONCE_SIZE)
        shared_key = random(SecretBox.KEY_SIZE)
        box = SecretBox(shared_key)
        plaintext = 'The ship has arrived at the port'
        encrypted_msg = box.encrypt(bytes(plaintext, 'utf-8'), nonce)
        nonce_str = b64encode(encrypted_msg.nonce).decode("utf-8")
        encrypted_msg_str = b64encode(encrypted_msg.ciphertext).decode("utf-8")
        message = f'{nonce_str}:{encrypted_msg_str}'

        decrypter = util.create_decrypter(shared_key)
        decrypted_text = decrypter(message)

        self.assertNotEqual(plaintext, encrypted_msg.ciphertext)
        self.assertEqual(plaintext, decrypted_text)

    def test_when_encrypting_with_a_key_the_decrypter_fails_when_not_using_the_shared_key(self):
        nonce = random(SecretBox.NONCE_SIZE)
        shared_key = random(SecretBox.KEY_SIZE)
        box = SecretBox(shared_key)
        plaintext = b'The ship has arrived at the port'
        encrypted_msg = box.encrypt(plaintext, nonce)
        nonce_str = b64encode(encrypted_msg.nonce).decode("utf-8")
        encrypted_msg_str = b64encode(encrypted_msg.ciphertext).decode("utf-8")
        message = f'{nonce_str}:{encrypted_msg_str}'

        incorrect_key = random(SecretBox.KEY_SIZE)
        decrypter = util.create_decrypter(incorrect_key)
        with self.assertRaises(Exception):
            decrypter(message)

        self.assertNotEqual(plaintext, encrypted_msg.ciphertext)



from requests import Session
import unittest
from unittest.mock import patch, MagicMock, mock_open, ANY

from harmony_service_lib import aws
from harmony_service_lib import util
from harmony_service_lib.http import request_context
from harmony_service_lib.message import Variable
from tests.util import config_fixture


class TestDownload(unittest.TestCase):
    def setUp(self):
        self.config = util.config(validate=False)

    @patch('harmony_service_lib.util.get_version')
    @patch('boto3.client')
    @patch('harmony_service_lib.aws.Config')
    def test_s3_download_sets_minimal_user_agent_on_boto_client(self, boto_cfg, client, get_version):
        fake_lib_version = '0.1.0'
        get_version.return_value = fake_lib_version
        cfg = config_fixture()
        boto_cfg_instance = MagicMock()
        boto_cfg.return_value = boto_cfg_instance
        with patch('builtins.open', mock_open()):
            util.download('s3://example/file.txt', 'tmp', access_token='', cfg=cfg)
        boto_cfg.assert_called_with(user_agent_extra=f'harmony (unknown version) harmony-service-lib/{fake_lib_version}')
        client.assert_called_with(service_name='s3', config=boto_cfg_instance, region_name=ANY)

    @patch('harmony_service_lib.util.get_version')
    @patch('boto3.client')
    @patch('harmony_service_lib.aws.Config')
    def test_s3_download_sets_harmony_user_agent_on_boto_client(self, boto_cfg, client, get_version):
        fake_lib_version = '0.1.0'
        get_version.return_value = fake_lib_version
        harmony_user_agt = 'harmony/3.3.3 (harmony-test)'
        cfg = config_fixture(user_agent=harmony_user_agt)
        boto_cfg_instance = MagicMock()
        boto_cfg.return_value = boto_cfg_instance
        with patch('builtins.open', mock_open()):
            util.download('s3://example/file.txt', 'tmp', access_token='', cfg=cfg)
        boto_cfg.assert_called_with(user_agent_extra=f'{harmony_user_agt} harmony-service-lib/{fake_lib_version}')
        client.assert_called_with(service_name='s3', config=boto_cfg_instance, region_name=ANY)

    @patch('harmony_service_lib.util.get_version')
    @patch('boto3.client')
    @patch('harmony_service_lib.aws.Config')
    def test_s3_download_sets_app_name_on_boto_client(self, boto_cfg, client, get_version):
        app_name = 'gdal-subsetter'
        fake_lib_version = '0.1.0'
        get_version.return_value = fake_lib_version
        cfg = config_fixture(app_name=app_name)
        boto_cfg_instance = MagicMock()
        boto_cfg.return_value = boto_cfg_instance
        with patch('builtins.open', mock_open()):
            util.download('s3://example/file.txt', 'tmp', access_token='', cfg=cfg)
        boto_cfg.assert_called_with(user_agent_extra=f'harmony (unknown version) harmony-service-lib/{fake_lib_version} ({app_name})')
        client.assert_called_with(service_name='s3', config=boto_cfg_instance, region_name=ANY)

    @patch('harmony_service_lib.util.get_version')
    @patch('harmony_service_lib.aws.download')
    @patch('harmony_service_lib.aws.Config')
    def test_s3_download_does_not_set_api_request_uuid(self, boto_cfg, aws_download, get_version):
        request_context['request_id'] = 'abc123'
        app_name = 'gdal-subsetter'
        fake_lib_version = '0.1.0'
        get_version.return_value = fake_lib_version
        cfg = config_fixture(app_name=app_name)
        boto_cfg_instance = MagicMock()
        boto_cfg.return_value = boto_cfg_instance
        with patch('builtins.open', mock_open()):
            util.download('s3://example/file.txt', 'tmp', access_token='', cfg=cfg)
        aws_download.assert_called_with(ANY, 's3://example/file.txt', ANY, ANY )

    @patch('harmony_service_lib.util.get_version')
    @patch.object(Session, 'get')
    def test_http_download_sets_api_request_uuid(self, get, get_version):
        request_context['request_id'] = 'abc123'
        app_name = 'gdal-subsetter'
        fake_lib_version = '0.1.0'
        get_version.return_value = fake_lib_version
        cfg = config_fixture(app_name=app_name)
        with patch('builtins.open', mock_open()):
            util.download('http://example/file.txt', 'tmp', access_token='', cfg=cfg)
        get.assert_called_with('http://example/file.txt?A-api-request-uuid=abc123',  headers={'user-agent': f'harmony (unknown version) harmony-service-lib/{fake_lib_version} (gdal-subsetter)'}, timeout=60, stream=True)

    @patch('harmony_service_lib.util.get_version')
    @patch.object(Session, 'get')
    def test_https_download_sets_api_request_uuid(self, get, get_version):
        request_context['request_id'] = 'abc123'
        app_name = 'gdal-subsetter'
        fake_lib_version = '0.1.0'
        get_version.return_value = fake_lib_version
        cfg = config_fixture(app_name=app_name)
        with patch('builtins.open', mock_open()):
            util.download('https://example/file.txt', 'tmp', access_token='', cfg=cfg)
        get.assert_called_with('https://example/file.txt?A-api-request-uuid=abc123',  headers={'user-agent': f'harmony (unknown version) harmony-service-lib/{fake_lib_version} (gdal-subsetter)'}, timeout=60, stream=True)

    @patch('harmony_service_lib.util.get_version')
    @patch.object(Session, 'post')
    def test_http_download_with_post_sets_api_request_uuid(self, post, get_version):
        request_context['request_id'] = 'abc123'
        app_name = 'gdal-subsetter'
        fake_lib_version = '0.1.0'
        get_version.return_value = fake_lib_version
        data = { 'foo': 'bar' }
        cfg = config_fixture(app_name=app_name)
        with patch('builtins.open', mock_open()):
            util.download('http://example/file.txt', 'tmp', access_token='', data=data, cfg=cfg)
        post.assert_called_with('http://example/file.txt?A-api-request-uuid=abc123',  headers={'user-agent': f'harmony (unknown version) harmony-service-lib/{fake_lib_version} (gdal-subsetter)', 'Content-Type': 'application/x-www-form-urlencoded'}, data = { 'foo': 'bar' }, timeout=60, stream=True)


    @patch('harmony_service_lib.util.get_version')
    @patch.object(Session, 'post')
    def test_https_download_with_post_sets_api_request_uuid(self, post, get_version):
        request_context['request_id'] = 'abc123'
        app_name = 'gdal-subsetter'
        fake_lib_version = '0.1.0'
        get_version.return_value = fake_lib_version
        data = { 'foo': 'bar' }
        cfg = config_fixture(app_name=app_name)
        with patch('builtins.open', mock_open()):
            util.download('https://example/file.txt', 'tmp', access_token='', data=data, cfg=cfg)
        post.assert_called_with('https://example/file.txt?A-api-request-uuid=abc123',  headers={'user-agent': f'harmony (unknown version) harmony-service-lib/{fake_lib_version} (gdal-subsetter)', 'Content-Type': 'application/x-www-form-urlencoded'}, data = { 'foo': 'bar' }, timeout=60, stream=True)


    @patch('harmony_service_lib.util.get_version')
    @patch.object(Session, 'post')
    def test_http_download_with_long_url_get_becomes_post(self, post, get_version):
        request_context['request_id'] = 'abc123'
        app_name = 'gdal-subsetter'
        fake_lib_version = '0.1.0'
        get_version.return_value = fake_lib_version
        # set post_url_length to 300 and download with url longer than 300, the download will be done with POST
        cfg = config_fixture(app_name=app_name,post_url_length=300)
        with patch('builtins.open', mock_open()):
            util.download('https://opendap.uat.earthdata.nasa.gov/collections/C1245618475-EEDTEST/granules/GPM_3IMERGHH.06:3B-HHR.MS.MRG.3IMERG.20200118-S233000-E235959.1410.V06B.HDF5.dap.nc4?dap4.ce=%2FGrid%2Ftime%3B%2FGrid%2Flon%3B%2FGrid%2Flat_bnds%3B%2FGrid%2Ftime_bnds%3B%2FGrid%2Flon_bnds%3B%2FGrid%2Flat',
            'tmp',
            access_token='',
            cfg=cfg)
        post.assert_called_with('https://opendap.uat.earthdata.nasa.gov/collections/C1245618475-EEDTEST/granules/GPM_3IMERGHH.06:3B-HHR.MS.MRG.3IMERG.20200118-S233000-E235959.1410.V06B.HDF5.dap.nc4',
        headers={'user-agent': f'harmony (unknown version) harmony-service-lib/{fake_lib_version} (gdal-subsetter)', 'Content-Type': 'application/x-www-form-urlencoded'}, data = 'dap4.ce=%2FGrid%2Ftime%3B%2FGrid%2Flon%3B%2FGrid%2Flat_bnds%3B%2FGrid%2Ftime_bnds%3B%2FGrid%2Flon_bnds%3B%2FGrid%2Flat&A-api-request-uuid=abc123', timeout=60, stream=True)


class TestStage(unittest.TestCase):
    def setUp(self):
        self.config = util.config(validate=False)

    @patch('boto3.client')
    def test_uploads_to_s3_and_returns_its_s3_url(self, client):
        # Sets a non-test ENV environment variable to force things through the (mocked) download path
        s3 = MagicMock()
        s3.generate_presigned_url.return_value = 'https://example.com/presigned.txt'
        client.return_value = s3
        cfg = config_fixture(use_localstack=True, staging_bucket='example', staging_path='staging/path')

        result = util.stage('file.txt', 'remote.txt', 'text/plain', cfg=cfg)

        s3.upload_file.assert_called_with('file.txt', 'example', 'staging/path/remote.txt',
                                          ExtraArgs={'ContentType': 'text/plain'})
        self.assertEqual(result, 's3://example/staging/path/remote.txt')

    @patch('boto3.client')
    def test_uses_location_prefix_when_provided(self, client):
        # Sets a non-test ENV environment variable to force things through the (mocked) download path
        s3 = MagicMock()
        s3.generate_presigned_url.return_value = 'https://example.com/presigned.txt'
        client.return_value = s3
        cfg = config_fixture(use_localstack=True, staging_bucket='example', staging_path='staging/path')

        result = util.stage('file.txt', 'remote.txt', 'text/plain',
                            location="s3://different-example/public/location/", cfg=cfg)

        s3.upload_file.assert_called_with('file.txt', 'different-example', 'public/location/remote.txt',
                                          ExtraArgs={'ContentType': 'text/plain'})
        self.assertEqual(result, 's3://different-example/public/location/remote.txt')


class TestS3Parameters(unittest.TestCase):
    def test_when_using_localstack_it_uses_localstack_host(self):
        use_localstack = True
        localstack_host = 'testhost'
        region = 'tatooine-desert-1'

        expected = {
            'endpoint_url': f'http://{localstack_host}:4566',
            'use_ssl': False,
            'aws_access_key_id': 'ACCESS_KEY',
            'aws_secret_access_key': 'SECRET_KEY',
            'region_name': f'{region}'
        }

        actual = aws.aws_parameters(use_localstack, localstack_host, region)
        self.assertDictEqual(expected, actual)

    def test_when_not_using_localstack_it_ignores_localstack_host(self):
        use_localstack = False
        localstack_host = 'localstack'
        region = 'westeros-north-3'

        expected = {
            'region_name': f'{region}'
        }

        actual = aws.aws_parameters(use_localstack, localstack_host, region)

        self.assertDictEqual(expected, actual)


class TestGenerateOutputFilename(unittest.TestCase):
    def test_includes_provided_suffixes_ext(self):
        """Ensure the correct combinations of regridded, subsetted and
        reformatted are included in the correct order, per the optional
        arguments to the function."""
        url = 'https://example.com/fake-path/abc.123.nc/?query=true'
        ext = 'zarr'

        # Basic cases
        variables = []
        with self.subTest('No suffix options'):
            self.assertEqual(
                util.generate_output_filename(url, ext),
                'abc.123.zarr'
            )
        with self.subTest('Only is_subsetted'):
            self.assertEqual(
                util.generate_output_filename(url, ext, is_subsetted=True),
                'abc.123_subsetted.zarr'
            )

        with self.subTest('Only is_regridded'):
            self.assertEqual(
                util.generate_output_filename(url, ext, is_regridded=True),
                'abc.123_regridded.zarr'
            )

        with self.subTest('Only is_reformatted'):
            self.assertEqual(
                util.generate_output_filename(url, ext, is_reformatted=True),
                'abc.123_reformatted.zarr'
            )

        with self.subTest('is_subsetted and is_regridded'):
            self.assertEqual(
                util.generate_output_filename(url, ext, is_subsetted=True, is_regridded=True),
                'abc.123_regridded_subsetted.zarr'
            )

        with self.subTest('is_subsetted, is_regridded with empty variables list'):
            self.assertEqual(
                util.generate_output_filename(
                    url, ext, variable_subset=variables, is_subsetted=True, is_regridded=True
                ),
                'abc.123_regridded_subsetted.zarr'
            )

        with self.subTest('is_subsetted and is_reformatted'):
            self.assertEqual(
                util.generate_output_filename(url, ext, is_subsetted=True, is_reformatted=True),
                'abc.123_subsetted_reformatted.zarr'
            )

        with self.subTest('is_regridded and is_reformatted'):
            self.assertEqual(
                util.generate_output_filename(url, ext, is_regridded=True, is_reformatted=True),
                'abc.123_regridded_reformatted.zarr'
            )

        with self.subTest('is_subsetted, is_regridded and is_reformatted'):
            self.assertEqual(
                util.generate_output_filename(
                    url, ext, is_subsetted=True, is_regridded=True, is_reformatted=True
                ),
                'abc.123_regridded_subsetted_reformatted.zarr'
            )

    def test_includes_single_variable_name_replacing_slashes(self):
        url = 'https://example.com/fake-path/abc.123.nc/?query=true'
        ext = 'zarr'

        # Variable name contains full path with '/' ('/' replaced with '_')
        variables = ['/path/to/VarB']
        self.assertEqual(
            util.generate_output_filename(url, ext, variable_subset=variables, is_subsetted=True, is_regridded=True),
            'abc.123_path_to_VarB_regridded_subsetted.zarr'
        )

    def test_decodes_encoded_chars(self):
        url = 'https://example.com/fake-path/GPM_3IMERGHH.06%5D3B-HHR.MS.MRG.3IMERG.20200101-S120000-E122959.0720.V06B.HDF5'
        self.assertEqual(
            util.generate_output_filename(url),
            'GPM_3IMERGHH.06]3B-HHR.MS.MRG.3IMERG.20200101-S120000-E122959.0720.V06B.HDF5'
        )

    def test_replaces_encoded_slash_with_underscore(self):
        url = 'https://example.com/fake-path/a/b/a%2fb%2F%2Fc.hdf5'
        self.assertEqual(
            util.generate_output_filename(url),
            'a_b_c.hdf5'
        )

    def test_runs_of_underscores_replaced_with_single(self):
        url = 'https://example.com/fake-path/granule__base___name.nc4'
        self.assertEqual(
            util.generate_output_filename(url, variable_subset=['/Grid/precipitationCal']),
            'granule_base_name_Grid_precipitationCal.nc4'
        )

    def test_leading_or_trailing_underscores_are_removed(self):
        url = 'https://example.com/fake-path/__granule__base___name.nc4__'
        self.assertEqual(
            util.generate_output_filename(url, variable_subset=['/Grid/precipitationCal']),
            'granule_base_name_Grid_precipitationCal.nc4'
        )

    def test_underscores_before_or_after_periods_are_removed(self):
        url = 'https://example.com/fake-path/__granule__base___name_.__nc4__'
        self.assertEqual(
            util.generate_output_filename(url, variable_subset=['/Grid/precipitationCal']),
            'granule_base_name_Grid_precipitationCal.nc4'
        )

    def test_replaces_colon_with_underscore(self):
        url = 'https://example.com/fake-path/q:q.nc4'
        self.assertEqual(
            util.generate_output_filename(url),
            'q_q.nc4'
        )
        url = 'https://example.com/fake-path/q%3Aq.nc4'
        self.assertEqual(
            util.generate_output_filename(url),
            'q_q.nc4'
        )
        url = 'https://example.com/fake-path/q%3Aq:q%2fq.nc4'
        self.assertEqual(
            util.generate_output_filename(url),
            'q_q_q_q.nc4'
        )

    def test_includes_single_variable(self):
        url = 'https://example.com/fake-path/abc.123.nc/?query=true'
        ext = 'zarr'

        # Single variable cases
        variables = ['VarA']
        self.assertEqual(
            util.generate_output_filename(url, ext),
            'abc.123.zarr'
        )
        self.assertEqual(
            util.generate_output_filename(url, ext, is_subsetted=True, is_regridded=True),
            'abc.123_regridded_subsetted.zarr'
        )
        self.assertEqual(
            util.generate_output_filename(url, ext, variable_subset=variables),
            'abc.123_VarA.zarr'
        )
        self.assertEqual(
            util.generate_output_filename(url, ext, variable_subset=variables, is_subsetted=True, is_regridded=True),
            'abc.123_VarA_regridded_subsetted.zarr'
        )
        self.assertEqual(
            util.generate_output_filename(
                url, ext, variable_subset=variables, is_subsetted=True, is_regridded=True, is_reformatted=True,
            ),
            'abc.123_VarA_regridded_subsetted_reformatted.zarr'
        )

    def test_excludes_multiple_variable(self):
        url = 'https://example.com/fake-path/abc.123.nc/?query=true'
        ext = 'zarr'

        # Multiple variable cases (no variable name in suffix)
        variables = ['VarA', 'VarB']
        self.assertEqual(
            util.generate_output_filename(url, ext, is_subsetted=True, is_regridded=True),
            'abc.123_regridded_subsetted.zarr'
        )
        self.assertEqual(
            util.generate_output_filename(url, ext, variable_subset=variables, is_subsetted=True, is_regridded=True),
            'abc.123_regridded_subsetted.zarr'
        )

    def test_avoids_overwriting_single_suffixes(self):
        ext = 'zarr'

        # URL already containing a suffix
        variables = ['VarA']
        url = 'https://example.com/fake-path/abc.123_regridded.zarr'
        self.assertEqual(
            util.generate_output_filename(url, ext, is_subsetted=True),
            'abc.123_regridded_subsetted.zarr'
        )
        self.assertEqual(
            util.generate_output_filename(url, ext, variable_subset=variables, is_subsetted=True, is_regridded=True),
            'abc.123_VarA_regridded_subsetted.zarr'
        )

    def test_avoids_overwriting_multiple_suffixes(self):
        ext = 'zarr'
        # URL already containing all suffixes
        variables = ['VarA']
        url = 'https://example.com/fake-path/abc.123_VarA_regridded_subsetted.zarr'
        self.assertEqual(
            util.generate_output_filename(url, ext, variable_subset=variables, is_subsetted=True, is_regridded=True),
            'abc.123_VarA_regridded_subsetted.zarr'
        )

    def test_allows_variable_objects(self):
        ext = 'zarr'
        # URL already containing all suffixes
        variables = [Variable({'name': 'VarA'})]
        url = 'https://example.com/fake-path/abc.123.zarr'
        self.assertEqual(
            util.generate_output_filename(url, ext, variable_subset=variables),
            'abc.123_VarA.zarr'
        )


class TestBboxToGeometry(unittest.TestCase):
    def test_provides_a_single_polygon_for_bboxes_not_crossing_the_antimeridian(self):
        self.assertEqual(
            util.bbox_to_geometry([100, 0, -100, 50]),
            {
                'type': 'MultiPolygon',
                'coordinates': [
                    [[[-180, 0], [-180, 50], [-100, 50], [-100, 0], [-180, 0]]],
                    [[[100, 0], [100, 50], [180, 50], [180, 0], [100, 0]]]
                ]
            })

    def test_splits_bboxes_that_cross_the_antimeridian(self):
        self.assertEqual(
            util.bbox_to_geometry([-100, 0, 100, 50]),
            {
                'type': 'Polygon',
                'coordinates': [
                    [[-100, 0], [-100, 50], [100, 50], [100, 0], [-100, 0]]
                ]
            })



import os
import unittest
from unittest.mock import patch

import harmony_service_lib.util
from harmony_service_lib import cli, BaseHarmonyAdapter
from pystac import Catalog
from tests.util import cli_test


class MockAdapter(BaseHarmonyAdapter):
    """
    Dummy class to mock adapter calls and record the input messages
    """
    messages = []
    errors = []
    cleaned_up = []
    result_catalog = Catalog(
        id='example id',
        description='An empty STAC catalog',
        stac_extensions=[]
    )

    def __init__(self, message, catalog=None):
        super().__init__(message, catalog)
        MockAdapter.messages.append(message.data)

    def invoke(self):
        return (self.message, self.result_catalog)

    def cleanup(self):
        MockAdapter.cleaned_up.append(True)


class TestIsHarmonyCli(unittest.TestCase):
    @cli_test('--something-else', 'invoke')
    def test_when_not_passing_harmony_action_it_returns_false(self, parser):
        parser.add_argument('--something-else')
        args = parser.parse_args()
        self.assertFalse(cli.is_harmony_cli(args))

    @cli_test('--harmony-action', 'invoke')
    def test_when_passing_harmony_action_it_returns_true(self, parser):
        args = parser.parse_args()
        self.assertTrue(cli.is_harmony_cli(args))

    @cli_test()
    def test_when_passing_nothing_it_returns_false(self, parser):
        args = parser.parse_args()
        self.assertFalse(cli.is_harmony_cli(args))


class TestCliInvokeAction(unittest.TestCase):
    def setUp(self):
        self.config = harmony_service_lib.util.config(validate=False)
        with open('/tmp/operation.json', 'w') as f:
            f.write('{"test": "input"}')

    def tearDown(self):
        os.remove('/tmp/operation.json')
        MockAdapter.messages = []
        MockAdapter.errors = []
        MockAdapter.cleaned_up = []

    @cli_test('--harmony-action', 'invoke')
    def test_when_harmony_input_is_not_provided_it_terminates_with_error(self, parser):
        with patch.object(parser, 'error') as error_method:
            args = parser.parse_args()
            cli.run_cli(parser, args, MockAdapter, self.config)
            error_method.assert_called_once_with(
                '--harmony-input or --harmony-input-file must be provided for --harmony-action=invoke')

    @cli_test('--harmony-action', 'invoke', '--harmony-input', '{"test": "input"}')
    def test_when_harmony_metadata_dir_is_not_provided_it_terminates_with_error(self, parser):
        with patch.object(parser, 'error') as error_method:
            args = parser.parse_args()
            cli.run_cli(parser, args, MockAdapter, self.config)
            error_method.assert_called_once_with(
                '--harmony-metadata-dir must be provided for --harmony-action=invoke')

    @cli_test('--harmony-action', 'invoke', '--harmony-input', '{"test": "input"}', '--harmony-metadata-dir', '/tmp')
    def test_when_harmony_input_is_provided_it_creates_and_invokes_an_adapter(self, parser):
        args = parser.parse_args()
        cli.run_cli(parser, args, MockAdapter, self.config)
        self.assertListEqual([{'test': 'input'}], MockAdapter.messages)

    @cli_test('--harmony-action', 'invoke', '--harmony-input-file', '/tmp/operation.json', '--harmony-metadata-dir', '/tmp')
    def test_when_harmony_input_file_is_provided_it_creates_and_invokes_an_adapter(self, parser):
        args = parser.parse_args()

        cli.run_cli(parser, args, MockAdapter, self.config)
        self.assertListEqual([{'test': 'input'}], MockAdapter.messages)

    @cli_test('--harmony-action', 'invoke', '--harmony-input', '{"test": "input"}', '--harmony-metadata-dir', '/tmp')
    def test_when_the_backend_service_throws_an_exception_after_response_it_does_not_respond_again(self, parser):
        class MockImpl(MockAdapter):
            def invoke(self):
                raise Exception('Something bad happened')

        args = parser.parse_args()
        try:
            cli.run_cli(parser, args, MockImpl, self.config)
        except Exception:
            pass
        self.assertListEqual(MockImpl.errors, [])

if __name__ == '__main__':
    unittest.main()



from copy import deepcopy
import unittest

from harmony_service_lib.message import Message
from .example_messages import minimal_message, minimal_source_message, full_message
from harmony_service_lib import util


class TestMessage(unittest.TestCase):
    def setUp(self):
        self.config = util.config(validate=False)

    def test_when_provided_a_full_message_it_parses_it_into_objects(self):
        message = Message(full_message)

        self.assertEqual(message.version, '0.22.0')
        self.assertEqual(message.callback, 'http://localhost/some-path')
        self.assertEqual(message.stagingLocation, 's3://example-bucket/public/some-org/some-service/some-uuid/')
        self.assertEqual(message.user, 'jdoe')
        self.assertEqual(message.client, 'curl')
        self.assertEqual(message.requestId, '00001111-2222-3333-4444-555566667777')
        self.assertEqual(message.isSynchronous, True)
        self.assertEqual(message.accessToken, 'ABCD1234567890')
        self.assertEqual(message.concatenate, True)
        self.assertEqual(message.average, 'time')
        self.assertEqual(message.sources[0].collection, 'C0001-EXAMPLE')
        self.assertEqual(message.sources[0].shortName, 'example_1_data')
        self.assertEqual(message.sources[0].versionId, '1')
        self.assertEqual(message.sources[0].variables[0].id, 'V0001-EXAMPLE')
        self.assertEqual(message.sources[0].variables[0].name, 'ExampleVar1')
        self.assertEqual(message.sources[0].variables[0].fullPath, 'example/path/ExampleVar1')
        self.assertEqual(message.sources[0].variables[0].type, 'SCIENCE_VARIABLE')
        self.assertEqual(message.sources[0].variables[0].subtype, 'SCIENCE_ARRAY')
        self.assertEqual(message.sources[0].variables[0].relatedUrls[0].urlContentType, 'DistributionURL')
        self.assertEqual(message.sources[0].variables[0].relatedUrls[0].type, 'GET DATA')
        self.assertEqual(message.sources[0].variables[0].relatedUrls[0].subtype, 'EOSDIS DATA POOL')
        self.assertEqual(message.sources[0].variables[0].relatedUrls[0].description, 'This URL points to some text data.')
        self.assertEqual(message.sources[0].variables[0].relatedUrls[0].url, 'http://example.com/file649.txt')
        self.assertEqual(message.sources[0].variables[0].relatedUrls[0].mimeType, 'text/plain')
        self.assertEqual(message.sources[0].variables[0].relatedUrls[0].format, 'ASCII')
        self.assertEqual(message.sources[0].variables[0].visualizations[0]['Name'], 'Test1234')
        self.assertEqual(message.sources[0].coordinateVariables[0].id, 'V1233801718-EEDTEST')
        self.assertEqual(message.sources[0].coordinateVariables[0].name, 'lat')
        self.assertEqual(message.sources[0].coordinateVariables[0].fullPath, 'lat')
        self.assertEqual(message.sources[0].coordinateVariables[0].type, 'COORDINATE')
        self.assertEqual(message.sources[0].coordinateVariables[0].subtype, 'LATITUDE')
        self.assertEqual(message.sources[0].granules[1].id, 'G0002-EXAMPLE')
        self.assertEqual(message.sources[0].granules[1].name, 'Example2')
        self.assertEqual(message.sources[0].granules[1].url, 'file://example/example_granule_2.txt')
        self.assertEqual(message.sources[0].granules[1].temporal.start, '2003-03-03T03:03:03Z')
        self.assertEqual(message.sources[0].granules[1].temporal.end, '2004-04-04T04:04:04Z')
        self.assertEqual(message.sources[0].granules[1].bbox, [-5, -6, 7, 8])
        self.assertEqual(message.sources[1].collection, 'C0002-EXAMPLE')
        self.assertEqual(message.sources[1].shortName, 'example_2_data')
        self.assertEqual(message.sources[1].versionId, '1')
        self.assertEqual(message.sources[1].variables[0].fullPath, 'example/path/ExampleVar2')
        self.assertEqual(message.sources[1].visualizations[0]['Name'], 'Test123')
        self.assertEqual(message.format.crs, 'CRS:84')
        self.assertEqual(message.format.srs.proj4, '+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs')
        self.assertEqual(message.format.srs.wkt, 'PROJCS[ ... ]')
        self.assertEqual(message.format.srs.epsg, 'EPSG:7030')
        self.assertEqual(message.format.isTransparent, True)
        self.assertEqual(message.format.mime, 'image/tiff')
        self.assertEqual(message.format.width, 800)
        self.assertEqual(message.format.height, 600)
        self.assertEqual(message.format.dpi, 72)
        self.assertEqual(message.format.interpolation, 'near')
        self.assertEqual(message.format.scaleExtent.x.min, 0.5)
        self.assertEqual(message.format.scaleExtent.x.max, 125)
        self.assertEqual(message.format.scaleExtent.y.min, 52)
        self.assertEqual(message.format.scaleExtent.y.max, 75.22)
        self.assertEqual(message.temporal.start, '1999-01-01T10:00:00Z')
        self.assertEqual(message.temporal.end, '2020-02-20T15:00:00Z')
        self.assertEqual(message.subset.bbox, [-91.1, -45.0, 91.1, 45.0])
        self.assertEqual(message.subset.shape.href, 's3://example-bucket/shapefiles/abcd.json')
        self.assertEqual(message.subset.shape.type, 'application/geo+json')
        self.assertEqual(message.subset.point, [-160.2, 80.2])
        self.assertEqual(message.subset.dimensions[0].name, 'XDim')
        self.assertEqual(message.subset.dimensions[0].min, 0.5)
        self.assertEqual(message.subset.dimensions[0].max, 12.0)
        self.assertEqual(message.subset.dimensions[1].name, 'YDim')
        self.assertEqual(message.subset.dimensions[1].min, None)
        self.assertEqual(message.subset.dimensions[1].max, 10)
        self.assertEqual(message.extendDimensions, ["lat", "lon"])
        self.assertEqual(message.pixelSubset, True)
        self.assertEqual(message.extraArgs['cut'], False)


    def test_when_provided_a_minimal_message_it_parses_it_into_objects(self):
        message = Message(minimal_message)

        self.assertEqual(message.version, '0.22.0')
        self.assertEqual(message.callback, 'http://localhost/some-path')
        self.assertEqual(message.stagingLocation, 's3://example-bucket/public/some-org/some-service/some-uuid/')
        self.assertEqual(message.user, 'jdoe')
        self.assertEqual(message.client, 'curl')
        self.assertEqual(message.requestId, '00001111-2222-3333-4444-555566667777')
        self.assertEqual(message.accessToken, 'ABCD1234567890')
        self.assertEqual(message.sources, [])
        self.assertEqual(message.format.crs, None)
        self.assertEqual(message.format.isTransparent, None)
        self.assertEqual(message.format.mime, None)
        self.assertEqual(message.format.width, None)
        self.assertEqual(message.format.height, None)
        self.assertEqual(message.format.dpi, None)
        self.assertEqual(message.subset.bbox, None)
        self.assertEqual(message.subset.dimensions, None)

    def test_when_provided_a_message_with_minimal_source_it_parses_it_into_objects(self):
        message = Message(minimal_source_message)

        self.assertEqual(message.version, '0.22.0')
        self.assertEqual(message.callback, 'http://localhost/some-path')
        self.assertEqual(message.user, 'jdoe')
        self.assertEqual(message.accessToken, 'ABCD1234567890')
        self.assertEqual(message.sources[0].collection, 'C0001-EXAMPLE')
        self.assertEqual(message.sources[0].shortName, 'example_1_data')
        self.assertEqual(message.sources[0].versionId, '1')
        self.assertEqual(message.sources[0].variables, [])
        self.assertEqual(message.sources[0].coordinateVariables, [])
        self.assertEqual(message.sources[0].granules, [])

    def test_granules_attribute_returns_all_child_granules(self):
        message = Message(full_message)

        self.assertEqual(len(message.granules), 4)
        self.assertEqual(message.granules[0].id, 'G0001-EXAMPLE')
        self.assertEqual(message.granules[1].id, 'G0002-EXAMPLE')
        self.assertEqual(message.granules[2].id, 'G0003-EXAMPLE')
        self.assertEqual(message.granules[3].id, 'G0004-EXAMPLE')

    def test_granules_link_to_their_parent_collection(self):
        message = Message(full_message)

        self.assertEqual(message.granules[0].collection, 'C0001-EXAMPLE')
        self.assertEqual(message.granules[1].collection, 'C0001-EXAMPLE')
        self.assertEqual(message.granules[2].collection, 'C0002-EXAMPLE')
        self.assertEqual(message.granules[3].collection, 'C0002-EXAMPLE')

    def test_granules_link_to_their_subset_variables(self):
        message = Message(full_message)

        self.assertEqual(message.granules[0].variables[0].id, 'V0001-EXAMPLE')
        self.assertEqual(message.granules[1].variables[0].id, 'V0001-EXAMPLE')
        self.assertEqual(message.granules[2].variables[0].id, 'V0002-EXAMPLE')
        self.assertEqual(message.granules[3].variables[0].id, 'V0002-EXAMPLE')

    def test_digest_returns_a_unique_string_per_message(self):
        message1 = Message(full_message)
        message2 = Message(minimal_source_message)
        message3 = Message(minimal_message)

        self.assertNotEqual(message1, message2)
        self.assertNotEqual(message2, message3)
        self.assertNotEqual(message3, message1)

    def test_processing_a_property_removes_it_from_json_output(self):
        message = Message(deepcopy(full_message))

        self.assertEqual(message.subset.process('bbox'), [-91.1, -45.0, 91.1, 45.0])
        self.assertEqual(message.format.process('interpolation'), 'near')
        message.sources[0].process('variables')

        # Retained in the original message
        self.assertEqual(message.subset.bbox, [-91.1, -45.0, 91.1, 45.0])
        self.assertEqual(message.format.interpolation, 'near')

        # Removed from the output
        output = Message(message.json)
        self.assertEqual(output.subset.bbox, None)
        self.assertEqual(output.format.interpolation, None)
        self.assertEqual(output.sources[0].variables, [])

        # Nearby properties are fine
        self.assertEqual(message.subset.shape.href, 's3://example-bucket/shapefiles/abcd.json')
        self.assertEqual(output.format.crs, 'CRS:84')
        self.assertEqual(message.format.srs.proj4, '+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs')
        self.assertEqual(output.sources[1].variables[0].fullPath, 'example/path/ExampleVar2')

        # Point property is generated
        self.assertEqual(message.subset.point, [-160.2, 80.2])

    def test_extra_args(self):
        message = Message(full_message)

        self.assertEqual(message.extraArgs['cut'], False)
        self.assertEqual(message.extraArgs['intParam'], 100)
        self.assertEqual(message.extraArgs['floatParam'], 123.456)
        self.assertEqual(message.extraArgs['stringParam'], 'value')
        self.assertEqual(message.extraArgs['arrayParam'], [1, 2, 3])
        self.assertEqual(message.extraArgs['objectParam'], {'name': 'obj1', 'attributes': ['x', 'y']})
        self.assertIsNone(message.extraArgs['nonExistent'])


import unittest
import copy
from io import StringIO

from harmony_service_lib.logging import build_logger
from tests.util import config_fixture
from harmony_service_lib.message import Message
from .example_messages import minimal_message


class TestLoggingRedaction(unittest.TestCase):

    def setUp(self):
        self.harmony_message = Message(minimal_message)
        self.token = self.harmony_message.accessToken
        self.buffer = StringIO()

    def configure_logger(self, text_logger):
        self.logger = build_logger(
            config_fixture(text_logger=text_logger), 
            stream=self.buffer)

    def test_msg_token_not_logged(self):
        self.configure_logger(text_logger=False)
        self.logger.info(self.harmony_message)
        log = self.buffer.getvalue()
        assert("accessToken = '<redacted>'" in log)
        assert(self.token not in log) 
        # check the same but with the text logger
        self.configure_logger(text_logger=True)
        self.logger.info(self.harmony_message)
        log = self.buffer.getvalue()
        assert("accessToken = '<redacted>'" in log)
        assert(self.token not in log)
        # check that the message wasn't mutated
        assert(self.harmony_message.accessToken == self.token)

    def test_arg_token_not_logged(self):
        log_call_arguments = ['the Harmony message is %s', self.harmony_message]
        self.configure_logger(text_logger=False)
        self.logger.info(*log_call_arguments)
        log = self.buffer.getvalue()
        assert("accessToken = '<redacted>'" in log)
        assert(self.token not in log)
        # check the same but with the text logger
        self.configure_logger(text_logger=True)
        self.logger.info(*log_call_arguments)
        log = self.buffer.getvalue()
        assert("accessToken = '<redacted>'" in log)
        assert(self.token not in log)
        # check that the message wasn't mutated
        assert(self.harmony_message.accessToken == self.token)

    def test_multiple_args_token_not_logged(self):
        log_call_arguments = ['the Harmony message is %s. Another arg: %s', self.harmony_message, "another arg"]
        self.configure_logger(text_logger=False)
        self.logger.info(*log_call_arguments)
        log = self.buffer.getvalue()
        assert("accessToken = '<redacted>'" in log)
        assert(self.token not in log)
        # check the same but with the text logger
        self.configure_logger(text_logger=True)
        self.logger.info(*log_call_arguments)
        log = self.buffer.getvalue()
        assert("accessToken = '<redacted>'" in log)
        assert(self.token not in log)
        # check that the message wasn't mutated
        assert(self.harmony_message.accessToken == self.token)

    def test_dict_token_not_logged(self):
        log_call_arguments = ['the Harmony message is %s', { 'the_harmony_message': self.harmony_message }]
        self.configure_logger(text_logger=False)
        self.logger.info(*log_call_arguments)
        log = self.buffer.getvalue()
        assert("accessToken = '<redacted>'" in log)
        assert(self.token not in log)
        # check the same but with the text logger
        self.configure_logger(text_logger=True)
        self.logger.info(*log_call_arguments)
        log = self.buffer.getvalue()
        assert("accessToken = '<redacted>'" in log)
        assert(self.token not in log)
        # check that the message wasn't mutated
        assert(self.harmony_message.accessToken == self.token)
        
        


"""
======
cli.py
======

Parses CLI arguments provided by Harmony and invokes the subsetter accordingly
"""

import json
import logging
from os import path, makedirs
import datetime

from pystac import Catalog, CatalogType
from pystac.layout import BestPracticesLayoutStrategy

from harmony_service_lib.exceptions import HarmonyException
from harmony_service_lib.message import Message
from harmony_service_lib.logging import setup_stdout_log_formatting, build_logger
from harmony_service_lib.util import (config, create_decrypter)
from harmony_service_lib.version import get_version
from harmony_service_lib.aws import is_s3, write_s3
from harmony_service_lib.s3_stac_io import S3StacIO


class MultiCatalogLayoutStrategy(BestPracticesLayoutStrategy):
    """
    Layout that adheres to what the Harmony server expects
    when multiple catalogs are output by a service.
    """

    def __init__(self, index):
        self.index = index

    def get_catalog_href(self, cat, parent_dir, is_root):
        """
        Returns the catalog href, using its index number as
        part of the file name, e.g. s3://outputs/catalog0.json.

        Parameters
        ----------
        parent_dir : string
            The parent directory of the catalog
        Returns
        -------
        The catalog href, postfixed with catalog{idx}.json
        """
        return path.join(parent_dir, f'catalog{self.index}.json')


def setup_cli(parser):
    """
    Adds Harmony arguments to the CLI being parsed by the provided parser

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The parser being used to parse CLI arguments
    """
    parser.add_argument('--harmony-action',
                        choices=['invoke'],
                        help=('the action Harmony needs to perform, "invoke" to run once and quit'))
    parser.add_argument('--harmony-input',
                        help=('the input data for the action provided by Harmony, required for '
                              '--harmony-action=invoke'))
    parser.add_argument('--harmony-input-file',
                        help=('the optional path to the input data for the action provided by Harmony'))
    parser.add_argument('--harmony-sources',
                        help=('file path that contains a STAC catalog with items and metadata to '
                              'be processed by the service.  Required for --harmony-action=invoke'))
    parser.add_argument('--harmony-metadata-dir',
                        help=('file path where output metadata should be written. The resulting '
                              'STAC catalog will be written to catalog.json in the supplied dir '
                              'with child resources in the same directory or a descendant '
                              'directory.  The remaining message, less any completed operations, '
                              'should be written to message.json in the supplied directory.  If '
                              'there is an error, it will be written to error.json in the supplied dir '))
    parser.add_argument('--harmony-data-location',
                        help=('the location where output data should be written, either a directory '
                              'or S3 URI prefix.  If set, overrides any value set by the message'))
    parser.add_argument('--harmony-visibility-timeout',
                        type=int,
                        default=600,
                        help=('the number of seconds the service is given to process a message '
                              'before processing is assumed to have failed'))
    parser.add_argument('--harmony-wrap-stdout',
                        action='store_const',
                        const=True,
                        help='Do not wrap STDOUT and STDERR in the Harmony log output format')


def is_harmony_cli(args):
    """
    Returns True if the passed parsed CLI arguments constitute a Harmony CLI invocation, False otherwise

    Parameters
    ----------
    args : Namespace
        Argument values parsed from the command line, presumably via ArgumentParser.parse_args

    Returns
    -------
    is_harmony_cli : bool
        True if the provided arguments constitute a Harmony CLI invocation, False otherwise
    """
    return args.harmony_action is not None


def _write_error(metadata_dir, message, category='Unknown', level='Error'):
    """
    Writes the given error message to error.json in the provided metadata dir

    Parameters
    ----------
    metadata_dir : string
        Directory into which the error should be written
    message : string
        The error message to write
    category : string
        The error category to write
    level : string
        The error level to write, can be 'Error' or 'Warning'. Default to 'Error'
    """
    error_data = {'error': message, 'category': category, 'level': level}
    if is_s3(metadata_dir):
        json_str = json.dumps(error_data)
        write_s3(f'{metadata_dir}error.json', json_str)
    else:
        with open(path.join(metadata_dir, 'error.json'), 'w') as file:
            json.dump(error_data, file)


def _build_adapter(AdapterClass, message_string, sources_path, data_location, config):
    """
    Creates the adapter to be invoked for the given harmony_service_lib input

    Parameters
    ----------
    AdapterClass : class
        The BaseHarmonyAdapter subclass to use to handle service invocations
    message_string : string
        The Harmony input message
    sources_path : string
        A file location containing a STAC catalog corresponding to the input message sources
    data_location : string
        The name of the directory where output should be written
    config : harmony_service_lib.util.Config
        A configuration instance for this service
    Returns
    -------
        BaseHarmonyAdapter subclass instance
            The adapter to be invoked
    """
    catalog = Catalog.from_file(sources_path) if bool(sources_path) else None

    secret_key = config.shared_secret_key

    if bool(secret_key):
        decrypter = create_decrypter(bytes(secret_key, 'utf-8'))
    else:
        def identity(arg):
            return arg
        decrypter = identity

    message = Message(json.loads(message_string), decrypter)
    if data_location:
        message.stagingLocation = data_location
    adapter = AdapterClass(message, catalog)
    adapter.set_config(config)

    return adapter


def _invoke(adapter, metadata_dir):
    """
    Handles --harmony-action=invoke by invoking the adapter for the given input message

    Parameters
    ----------
    adapter : BaseHarmonyAdapter
        The BaseHarmonyAdapter subclass to use to handle service invocations
    metadata_dir : string
        The name of the directory where STAC and message output should be written
    Returns
    -------
    True if the operation completed successfully, False otherwise
    """
    try:
        logging.info(f'Invoking adapter with harmony-service-lib-py version {get_version()}')
        s3_io = S3StacIO()
        is_s3_metadata_dir = is_s3(metadata_dir)
        if not is_s3_metadata_dir:
            makedirs(metadata_dir, exist_ok=True)
        (out_message, stac_output) = adapter.invoke()
        if isinstance(stac_output, list):
            for idx, catalog in enumerate(stac_output):
                catalog.normalize_and_save(metadata_dir, CatalogType.SELF_CONTAINED, MultiCatalogLayoutStrategy(idx))
            json_str = json.dumps([f'catalog{i}.json' for i, c in enumerate(stac_output)])
            s3_io.write_text(path.join(metadata_dir, 'batch-catalogs.json'), json_str)
            s3_io.write_text(path.join(metadata_dir, 'batch-count.txt'), f'{len(stac_output)}')
        else:  # assume stac_output is a single catalog
            stac_output.normalize_and_save(metadata_dir, CatalogType.SELF_CONTAINED)

        if not is_s3_metadata_dir:
            with open(path.join(metadata_dir, 'message.json'), 'w') as file:
                json.dump(out_message.output_data, file)
    except HarmonyException as err:
        logging.error(err, exc_info=1)
        _write_error(metadata_dir, err.message, err.category, err.level)
        raise
    except BaseException as err:
        logging.error(err, exc_info=1)
        _write_error(metadata_dir, 'Service request failed with an unknown error')
        raise


def run_cli(parser, args, AdapterClass, cfg=None):
    """
    Runs the Harmony CLI invocation captured by the given args

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The parser being used to parse CLI arguments, used to provide CLI argument errors
    args : Namespace
        Argument values parsed from the command line, presumably via ArgumentParser.parse_args
    AdapterClass : class
        The BaseHarmonyAdapter subclass to use to handle service invocations
    cfg : harmony_service_lib.util.Config
        A configuration instance for this service
    """
    if cfg is None:
        cfg = config()
    if args.harmony_wrap_stdout:
        setup_stdout_log_formatting(cfg)

    # read in the operation file passed in with --harmony-input-file if any
    if bool(args.harmony_input_file):
        with open(args.harmony_input_file, 'r') as f:
            args.harmony_input = f.read()

    if args.harmony_action == 'invoke':
        start_time = datetime.datetime.now()
        if not bool(args.harmony_input):
            parser.error(
                '--harmony-input or --harmony-input-file must be provided for --harmony-action=invoke')
        elif not bool(args.harmony_metadata_dir):
            parser.error(
                '--harmony-metadata-dir must be provided for --harmony-action=invoke')
        else:
            adapter = None
            try:
                adapter = _build_adapter(AdapterClass,
                                         args.harmony_input,
                                         args.harmony_sources,
                                         args.harmony_data_location,
                                         cfg)
                adapter.logger.info(f'timing.{cfg.app_name}.start')
                _invoke(adapter, args.harmony_metadata_dir)
            finally:
                time_diff = datetime.datetime.now() - start_time
                duration_ms = int(round(time_diff.total_seconds() * 1000))
                duration_logger = build_logger(cfg)
                extra_fields = {
                    'user': (
                        adapter.message.user
                        if adapter and adapter.message and hasattr(adapter.message, "user")
                        else ''
                    ),
                    'requestId': (
                        adapter.message.requestId
                        if adapter and adapter.message and hasattr(adapter.message, "requestId")
                        else ''
                    ),
                    'durationMs': duration_ms
                }
                duration_logger.info(f'timing.{cfg.app_name}.end', extra=extra_fields)



"""
This module includes various AWS-specific functions to stage data in S3.

This module relies on the harmony_service_lib.util.config and its environment variables to be
set for correct operation. See that module and the project README for details.
"""
from urllib.parse import urlparse
from os import environ
import boto3
from botocore.config import Config
from harmony_service_lib import util


def is_s3(url: str) -> bool:
    """Predicate to determine if a url is an S3 endpoint."""
    return url is not None and url.lower().startswith('s3')


def aws_parameters(use_localstack, localstack_host, region):
    """Constructs a configuration dict that can be used to create an aws client.

    Parameters
    ----------
    use_localstack : bool
        Whether to use the localstack in this environment.
    localstack_host : str
        The hostname of the localstack services (if use_localstack enabled).
    region : str
        The AWS region to connect to.
    Returns
    -------

    """
    if use_localstack:
        return {
            'endpoint_url': f'http://{localstack_host}:4566',
            'use_ssl': False,
            'aws_access_key_id': 'ACCESS_KEY',
            'aws_secret_access_key': 'SECRET_KEY',
            'region_name': region
        }
    else:
        return {
            'region_name': region
        }


def write_s3(url, txt):
    """
    Writes text to the given  s3 url.

    Parameters
    ----------
    url: The s3 file url.
    txt: The file contents.
    """
    parsed = urlparse(url)
    config = util.config(validate=environ.get('ENV') != 'test')
    service_params = aws_parameters(
            config.use_localstack, config.localstack_host, config.aws_default_region)
    bucket = parsed.netloc
    key = parsed.path[1:]
    s3 = boto3.resource("s3", **service_params)
    s3.Object(bucket, key).put(Body=txt)


def _get_aws_client(config, service, user_agent=None):
    """
    Returns a boto3 client for accessing the provided service.  Accesses the service in us-west-2
    unless "AWS_DEFAULT_REGION" is set.  If the environment variable "USE_LOCALSTACK" is set to "true",
    it will return a client that will access a LocalStack instance instead of AWS.

    Parameters
    ----------
    config : harmony_service_lib.util.Config
        The configuration for the current runtime environment.
    service : string
        The AWS service name for which to construct a client, e.g. "s3"
    user_agent : string
        The user agent that is requesting the aws service.
        E.g. harmony/0.0.0 (harmony-sit) harmony-service-lib/4.0 (gdal-subsetter)

    Returns
    -------
    s3_client : boto3.*.Client
        A client appropriate for accessing the provided service
    """
    boto_cfg = Config(user_agent_extra=user_agent)
    service_params = aws_parameters(config.use_localstack, config.localstack_host, config.aws_default_region)

    return boto3.client(service_name=service, config=boto_cfg, **service_params)


def download(config, url, destination_file, user_agent=None):
    """Download an S3 object to the specified destination directory.

    Parameters
    ----------
    config : harmony_service_lib.util.Config
        The configuration for the current runtime environment.
    destination_file : file-like
        The destination file where the object will be written. Must be
        a file-like object opened for binary write.
    user_agent : string
        The user agent that is requesting the download.
        E.g. harmony/0.0.0 (harmony-sit) harmony-service-lib/4.0 (gdal-subsetter)
    """
    bucket = url.split('/')[2]
    key = '/'.join(url.split('/')[3:])
    aws_client = _get_aws_client(config, 's3', user_agent)
    aws_client.download_fileobj(bucket, key, destination_file)


def stage(config, local_filename, remote_filename, mime, logger, location=None):
    """
    Stages the given local filename, including directory path, to an S3 location with the given
    filename and mime-type

    Requires the following environment variables:
        AWS_DEFAULT_REGION: The AWS region in which the S3 client is operating

    Parameters
    ----------
    config : harmony_service_lib.util.Config
        The configuration for the current runtime environment.
    local_filename : string
        A path and filename to the local file that should be staged
    remote_filename : string
        The basename to give to the remote file
    mime : string
        The mime type to apply to the staged file for use when it is served, e.g. "application/x-netcdf4"
    location : string
        The S3 prefix URL under which to place the output file.  If not provided, STAGING_BUCKET and
        STAGING_PATH must be set in the environment
    logger : logging
        The logger to use

    Returns
    -------
    url : string
        An s3:// URL to the staged file
    """
    key = None
    staging_bucket = config.staging_bucket

    if location is None:
        if config.staging_path:
            key = '%s/%s' % (config.staging_path, remote_filename)
        else:
            key = remote_filename
    else:
        _, _, staging_bucket, staging_path = location.split('/', 3)
        key = staging_path + remote_filename

    if config.env in ['dev', 'test'] and not config.use_localstack:
        logger.warning(f"ENV={config.env}"
                       f" and not using localstack, so we will not stage {local_filename} to {key}")
        return "http://example.com/" + key

    s3 = _get_aws_client(config, 's3')
    s3.upload_file(local_filename, staging_bucket, key, ExtraArgs={'ContentType': mime})

    return 's3://%s/%s' % (staging_bucket, key)



"""
==========
message.py
==========

Harmony message parsing and helper objects.  Callers should generally
only construct the 'Message' object and allow its children to be built
from the message JSON.
"""

import hashlib
import json
import copy
from warnings import warn


class JsonObject(object):
    """
    Base class for deserialized Harmony message objects

    Attributes
    ----------
    data : dictionary
        The JSON data object / dictionary used to build this object

    properties: list
        A list of properties that are included in string representations
    """
    reprdepth = 0

    def __init__(self, data, properties=[], list_properties={}):
        """
        Constructor

        Parameters
        ----------
        data : dictionary
            The JSON dictionary created by json.loads at the root of this object
        properties : list, optional
            A list of properties that should be extracted to attributes, by default []
        list_properties : dict, optional
            A dictionary of property name to type for properties that are lists of
            JSONObject classes, by default {}
        """
        self.output_data = data or {}
        self.data = copy.deepcopy(data) or {}
        self.properties = properties + list(list_properties.keys())
        self.processed = []
        for prop in properties:
            setattr(self, prop, data.get(prop))
        for prop in list_properties:
            Class = list_properties[prop]
            items = data.get(prop) or []
            value = [Class(item) for item in items]
            setattr(self, prop, value)

    def __getitem__(self, key):
        """
        Retrieve the value corresponding to a key in data

        Parameters
        ----------
        key : str
            The key to retrieve the value for

        Returns
        -------
        value : object or None
            The value corresponding to the key if it exists, otherwise None
        """
        return self.data.get(key)

    def process(self, *prop):
        """
        Marks the given property as having been processed and returns its value.
        If multiple properties are passed, returns their values an array

        Parameters
        ----------
        prop : string
            the name of the property having been processed

        Returns
        -------
        object
            the value of the property supplied
        """
        result = []
        for p in prop:
            self.output_data.pop(p, None)
            result.append(getattr(self, p))
        if len(result) == 1:
            return result[0]
        return result

    def __repr__(self):
        """
        Returns
        -------
        string
            A string representation of the object
        """
        result = ''
        JsonObject.reprdepth += 1
        try:
            spaces = '    ' * JsonObject.reprdepth
            result += '<' + self.__class__.__name__ + '\n'
            result += '\n'.join(["%s%s = %s" % (spaces, p, repr(getattr(self, p)))
                                 for p in self.properties])
            result += '>'
        finally:
            JsonObject.reprdepth -= 1
        return result


class Source(JsonObject):
    """
    A collection / granule / variable / coordinateVariable / visualization data source as found in
    the Harmony message "sources" list.

    Attributes
    ----------
    collection : string
        The id of the collection the data source's variables and granules are in
    shortName : string
        The unique short name of the collection as returned by the CMR
    versionId : string
        The version id of the collection as returned by the CMR
    variables : list
        A list of Variable objects for the variables which should be transformed
    coordinateVariables: list
        A list of Variable objects containing the coordinate variables for the
        collection.
    visualizations: list
        A list of objects containing the UMM-Vis data for the collection. This list will be
        empty if the user specifies any variables for subsetting.
    granules : list
        A list of Granule objects for the granules which should be operated on
    """

    def __init__(self, message_data):
        """
        Constructor

        Parameters
        ----------
        message_data : dictionary
            The Harmony message "sources" item to deserialize
        """
        super().__init__(message_data,
                         properties=['collection', 'shortName', 'versionId'],
                         list_properties={
                             'variables': Variable,
                             'coordinateVariables': Variable,
                             'visualizations': JsonObject,
                             'granules': Granule
                            }
                         )
        for granule in self.granules:
            granule.collection = self.collection
            granule.variables = self.variables


class Variable(JsonObject):
    """
    A data variable as found in a Harmony source's "variables" list

    Attributes
    ----------
    id : string
        The UMM-Var ID of the variable.
    name : string
        The UMM-Var short name of the variable, typically identifies layer name found in the
        science data file.
    fullPath : string
         The variable's absolute path within the file, including hierarchy.  Derived from
         UMM-Var group path combined with name.
    relatedUrls : list
         A list of RelatedUrl(s) for the variable.
    visualizations: list
        A list of Umm-Vis objects for the variable.
    """

    def __init__(self, message_data):
        """
        Constructor

        Parameters
        ----------
        message_data : dictionary
            The Harmony message "variables" or "coordinateVariables" item to deserialize
        """
        super().__init__(
            message_data,
            properties=['id', 'name', 'fullPath', 'type', 'subtype'],
            list_properties={
                'relatedUrls': RelatedUrl,
                'visualizations': JsonObject
            })


class RelatedUrl(JsonObject):
    """
    A related URL describes an external resource or location on the web
    (data access location, project home page, relevant software packages, etc.)

    Attributes
    ----------
    url : string
        Points to the location of the resource described by the RelatedUrl.
    urlContentType : string
        A keyword which describes the content of a link at a high level.
    type : string
        A keyword which specifies the content of a link.
    subtype : string
        A keyword which further specifies the content of a link.
    description : string
        Explains where the link navigates and the type of information it contains.
    format : string
        The format of the data.
    mimeType : string
        The mime type of the data.
    """

    def __init__(self, message_data):
        """
        Constructor

        Parameters
        ----------
        message_data : dictionary
            The Harmony message "relatedUrls" item to deserialize
        """
        super().__init__(message_data, properties=[
            'url', 'urlContentType', 'type', 'subtype', 'description', 'format', 'mimeType']
        )


class Granule(JsonObject):
    """
    A science granule as found in a Harmony source's "granules" list

    Attributes
    ----------
    id : string
        The CMR Granule ID of the granule
    name: string
        The granule's short name
    url: string
        The URL to the granule, preferentially an S3 URL.  Potentially behind EDL
    bbox : list
        A list of 4 floating point values corresponding to [West, South, East, North]
        coordinates of the granule's spatial MBR
    temporal: Temporal
        The temporal extent of the granule
    """

    def __init__(self, message_data):
        """
        Constructor

        Parameters
        ----------
        message_data : dictionary
            The Harmony message "granules" item to deserialize
        """
        super().__init__(message_data, properties=[
            'id', 'name', 'url', 'bbox', 'temporal'])
        warn('message.Granule is deprecated.  New workflows will use STAC catalogs instead',
             DeprecationWarning, stacklevel=2)
        self.local_filename = None
        self.collection = None
        self.variables = []
        if self.temporal is not None:
            self.temporal = Temporal(message_data['temporal'])


class MinMax(JsonObject):
    """
    Min and max parameters as found in a Harmony message's "format.scaleExtent.[x|y]" objects

    Attributes
    ----------
    min: float
        The min value for the attribute
    max: float
        The max value for the attribute
    """

    def __init__(self, message_data):
        """
        Constructor

        Parameters
        ----------
        message_data : dictionary
            The Harmony message "format.scaleExtent.[x|y]" object to deserialize
        """
        super().__init__(message_data, properties=['min', 'max'])


class Dimension(JsonObject):
    """
    Dimension subset parameters as found in a single dimension from the list of the
    Harmony message's "subset.dimensions" field.

    Attributes
    ----------
    name: string
        The name of the dimension
    min: float
        The min value for the dimension
    max: float
        The max value for the dimension
    """

    def __init__(self, message_data):
        """
        Constructor

        Parameters
        ----------
        message_data : dictionary
            A single dimension from the list of the Harmony message's "subset.dimensions"
            field
        """
        super().__init__(message_data, properties=['name', 'min', 'max'])


class ScaleExtent(JsonObject):
    """
    Scale extent parameters as found in a Harmony message's "format.scaleExtent" object

    Attributes
    ----------
    x: message.MinMax
        The min and max values for the scale extent for the X dimension
    y: message.MinMax
        The min and max values for the scale extent for the Y dimension
    """

    def __init__(self, message_data):
        """
        Constructor

        Parameters
        ----------
        message_data : dictionary
            The Harmony message "format.scaleExtent" object to deserialize
        """
        super().__init__(message_data, properties=['x', 'y'])
        if self.x is not None:
            self.x = MinMax(message_data['x'])
        if self.y is not None:
            self.y = MinMax(message_data['y'])


class ScaleSize(JsonObject):
    """
    Scale size parameters as found in a Harmony message's "format.scaleSize" object

    Attributes
    ----------
    x: float
        The scale size for the X dimension
    y: float
        The scale size for the Y dimension
    """

    def __init__(self, message_data):
        """
        Constructor

        Parameters
        ----------
        message_data : dictionary
            The Harmony message "format.scaleExtent" object to deserialize
        """
        super().__init__(message_data, properties=['x', 'y'])


class SRS(JsonObject):
    """
    Output CRS information as found in a Harmony message's "format.srs" object

    Attributes
    ----------
    proj4 : string
        The Proj4 representation for output CRS.
    wkt : string
        The WKT information for output CRS.
    epsg : string
        The EPSG designation, if suppplied or derived, for output CRS.
    """

    def __init__(self, message_data):
        """
        Constructor

        Parameters
        ----------
        message_data : dictionary
            The Harmony message "variables" item to deserialize
        """
        super().__init__(message_data, properties=['proj4', 'wkt', 'epsg'])


class Format(JsonObject):
    """
    Output format parameters as found in a Harmony message's "format" object

    Attributes
    ----------
    crs: string
        A proj4 string or EPSG code corresponding to the desired output projection
    srs: message.SRS
        The output CRS information; overlappting information with 'crs'
    isTransparent: boolean
        A boolean corresponding to whether or not nodata values should be set to transparent
        in the output if the file format allows it
    mime: string
        The mime type of the desired output file
    width: integer
        The pixel width of the desired output
    height: integer
        The pixel height of the desired output
    dpi: integer
        The number of pixels per inch in the desired output file, for image output formats
        that support it
    interpolation: string
        The interpolation method
    scaleExtent: message.ScaleExtent
        The scale extent in the x and y dimensions
    scaleSize: message.ScaleSize
        The scale size in the x and y dimensions
    """

    def __init__(self, message_data):
        """
        Constructor

        Parameters
        ----------
        message_data : dictionary
            The Harmony message "format" object to deserialize
        """
        super().__init__(message_data, properties=[
            'crs',
            'srs',
            'isTransparent',
            'mime',
            'width',
            'height',
            'dpi',
            'interpolation',
            'scaleExtent',
            'scaleSize'
        ])
        if self.srs is not None:
            self.srs = SRS(message_data['srs'])
        if self.scaleExtent is not None:
            self.scaleExtent = ScaleExtent(message_data['scaleExtent'])
        if self.scaleSize is not None:
            self.scaleSize = ScaleSize(message_data['scaleSize'])


class RemoteResource(JsonObject):
    """
    Remote resource

    Attributes
    ----------
    uri : string
        A string of the remote resource location
    type : string
        The resource's content type
    """

    def __init__(self, message_data):
        """
        Constructor

        Parameters
        ----------
        message_data : dictionary
            The Harmony message remote resource object to deserialize
        """
        super().__init__(message_data, properties=['href', 'type'])


class Subset(JsonObject):
    """
    Subsetting parameters as found in a Harmony message's "subset" object

    Attributes
    ----------
    bbox : list
        A list of 4 floating point values corresponding to [West, South, East, North]
        coordinates
    point: list containing 2 floating point values corresponding to longitude and latitude
    shape: RemoteResource
        A reference to a location containing a shapefile
    dimensions: list
        A list of Dimension objects to subset against
    """

    def __init__(self, message_data):
        """
        Constructor

        Parameters
        ----------
        message_data : dictionary
            The Harmony message "subset" object to deserialize
        """
        super().__init__(message_data, properties=['bbox', 'point', 'shape', 'dimensions'])
        if self.shape is not None:
            self.shape = RemoteResource(message_data['shape'])
        if self.dimensions is not None:
            dimensions = []
            for dimension in self.dimensions:
                dimensions.append(Dimension(dimension))
            self.dimensions = dimensions


class Temporal(JsonObject):
    """
    Temporal subsetting parameters as found in a Harmony message's "temporal" object

    Attributes
    ----------
    start : string
        An ISO 8601 datetime string for the earliest time for temporal subsetting
    end : string
        An ISO 8601 datetime string for the latest time for temporal subsetting
    """

    def __init__(self, message_data=None, start=None, end=None):
        """
        Constructor

        Parameters
        ----------
        message_data : dictionary, optional
            The Harmony message "temporal" object to deserialize
        start: string, optional
            The temporal range start as RFC-3339 date/time string
        start: end, optional
            The temporal range end as RFC-3339 date/time string
        """
        super().__init__(message_data or {}, properties=['start', 'end'])
        if start is not None:
            self.start = start
        if end is not None:
            self.end = end


class ExtraArgs(JsonObject):
    """
    Extra Args parameters as found in a Harmony message's "extraArgs" object
    The value of extra args parameter can be retrieved via ['<parameter>'],
    e.g. message.extraArgs['cut'] will return the value of 'cut' parameter in extraArgs.
    """

    def __init__(self, message_data):
        """
        Constructor

        Parameters
        ----------
        message_data : dictionary
            The Harmony message "extraArgs" object to deserialize
        """
        super().__init__(message_data)


class Message(JsonObject):
    """
    Top-level object corresponding to an incoming Harmony message.  Constructing
    this with a JSON string will deserialize the message into native Python object,
    perform any necessary version interpretation, and add some helpers to make access
    easier.  Generally, this object should be created and allowed to produce its
    child objects rather than directly instantiating Subset, Format, etc objects.
    For maximum compatibility and ease of use, services should prefer using objects
    of this class and their children rather than parsing Harmony's JSON.

    Attributes
    ----------
    version : string
        The semantic version of the Harmony message contained in the provided JSON
    callback : string
        (Deprecated) The URL that services must POST to when their execution is complete.
    stagingLocation : string
        An object store (S3) URL prefix under which services may elect to put their output.
        Services must have write access to the Harmony staging bucket for the deployed
        environment to use this value.  The location will be unique per Harmony request
        but services are responsible for ensuring no name clashes occur within a single
        request.  The prefix will end in a "/" character.
    isSynchronous : bool
        True if a user is awaiting an immediate response, False if the user is expecting
        the service to be performed at a later point.  This may influence prioritization
        of the request and impacts ability to send multi-file responses
    user : string
        The username of the user requesting the service.  If the message is coming from
        Harmony, services can assume that the provided username has been authenticated
    accessToken : string
        The Earthdata Login token for the caller. If present, the token is used as the
        identity for HTTP downloads.
    client : string
        A string indicating the client accessing the service, usually the harmony
        environment, e.g. "harmony-sit"
    requestId : string
        A UUID identifying the originating user request.  This should only be used for
        logging and tracing purposes, as a single user request may produce multiple
        service invocations.
    format: message.Format
        The Harmony message's output parameters
    subset: message.Subset
        The Harmony message's subsetting parameters
    temporal: message.Temporal
        The Harmony message's temporal subsetting parameters
    concatenate: bool
        True if the service should concatenate multiple input files into a single output
        file and false otherwise.
    average: string
        The averaging method to use
    extendDimensions: list
        A list of dimensions to extend.
    pixelSubset : bool
        True if pixel subset should be performed by the service.
    extraArgs: object
        A map of key (string type) and value (any type) pairs indicating the extra arguments
        that should be passed to the worker command
    """

    def __init__(self, json_str_or_dict, decrypter=lambda x: x):
        """
        Builds a Message object and all of its child objects by deserializing the
        provided JSON string and performing any necessary version interpretation.

        Parameters
        ----------
        json_str_or_dict : string | Object
            The incoming Harmony message as a JSON string or dict as parsed by `json.load()`
        decrypter : function
            A function that takes an encrypted value and returns it decrypted
        """

        if isinstance(json_str_or_dict, str):
            json_obj = json.loads(json_str_or_dict)
        else:
            json_obj = copy.deepcopy(json_str_or_dict)

        super().__init__(
            json_obj,
            properties=[
                'version',
                'callback',
                'stagingLocation',
                'isSynchronous',
                'user',
                'accessToken',
                'client',
                'requestId',
                'format',
                'subset',
                'temporal',
                'concatenate',
                'average',
                'extendDimensions',
                'pixelSubset',
                'extraArgs'
            ],
            list_properties={'sources': Source}
        )

        self.decrypter = decrypter

        if self.format is not None:
            self.format = Format(json_obj['format'])
        if self.subset is not None:
            self.subset = Subset(json_obj['subset'])
        if self.temporal is not None:
            self.temporal = Temporal(json_obj['temporal'])
        if self.accessToken is not None:
            self.accessToken = self.decrypter(self.accessToken)
        if self.extraArgs is not None:
            self.extraArgs = ExtraArgs(json_obj['extraArgs'])

    @property
    def json(self):
        return json.dumps(self.output_data)

    def digest(self):
        """
        Returns a shasum of the message, useful in providing unique output IDs

        Returns
        -------
        string
            The shasum of the message
        """
        return hashlib.sha256(self.json.encode('utf-8')).hexdigest()

    @property
    def granules(self):
        """
        A list of all the granules in all of the data sources.  Each granule
        links back to its source collection and requested variables, so it
        can be more convenient to use this granules list than to traverse
        the data sources themselves if services process granules individually

        Returns
        -------
        list
            A list of Granule objects for all of the granules in the message
        """
        result = []
        for source in self.sources:
            result += source.granules
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# For a fully annotated version of this file and what it does, see
# https://github.com/pypa/sampleproject/blob/master/setup.py

# To upload this file to PyPI you must build it then upload it:
# python setup.py sdist bdist_wheel  # build in 'dist' folder
# python-m twine upload dist/*  # 'twine' must be installed: 'pip install twine'
#
# 2024-09-24 - Updated to use pyproject.toml.

from setuptools import setup

setup()



"""Utilities for acting on Harmony Messages.

These are a collection of useful routines for validation and interrogation of
harmony_service_lib Messages.
"""

from typing import Any, List

from harmony_service_lib.message import Message


def has_self_consistent_grid(message: Message, allow_incomplete_grid: bool = False) -> bool:
    """Check the input Harmony message defines a self-consistent grid.

    At minimum a self-consistent grid should define the scale extents
    (minimum and maximum values) in the horizontal spatial dimensions and
    one of the following two pieces of information:

        * Message.format.scaleSize - defining the x and y pixel size.
        * Message.format.height and Message.format.width - the number of pixels
          in the x and y dimension.

    If all three pieces of information are supplied, they will be checked to
    ensure they are consistent with one another.

    If scaleExtent and scaleSize are defined, along with only one of height or
    width, the grid will be considered consistent if the three values for
    scaleExtent, scaleSize and specified dimension length, height or width, are
    consistent.

    If no grid parameters are provided, or only one of the three are defined,
    then the function will return the value of `allow_incomplete_grid`, as there is
    insufficient information to determine if the grid is self-consistent.

    Parameters
    ----------
        message : harmony_service_lib.message.Message
            The Harmony message object provided to a service for a request.
        allow_incomplete_grid : bool, optional
            Optional parameter stating whether the validation check should pass
            if the message does not contain any grid parameters. Applicable to
            instances when only a target projection is specified in a request,
            with the expectation that the target grid will cover the horizontal
            spatial area of the input granule. Default value is `False`.

    Returns
    -------
        bool
            Value indicating if the Harmony message parameters met the criteria
            for grid self-consistency. If there are no grid parameters, or only
            one of scaleExtents, scaleSize or height/width are provided, then
            the return value is determined by `allow_incomplete_grid`, which
            defaults to `False`.

    """
    if (
        has_scale_extents(message) and has_scale_sizes(message)
        and has_dimensions(message)
    ):
        consistent_grid = (_has_consistent_dimension(message, 'x')
                           and _has_consistent_dimension(message, 'y'))
    elif (
        has_scale_extents(message) and has_scale_sizes(message)
        and rgetattr(message, 'format.height') is not None
    ):
        consistent_grid = _has_consistent_dimension(message, 'y')
    elif (
        has_scale_extents(message) and has_scale_sizes(message)
        and rgetattr(message, 'format.width') is not None
    ):
        consistent_grid = _has_consistent_dimension(message, 'x')
    elif (
        has_scale_extents(message)
        and (has_scale_sizes(message) or has_dimensions(message))
    ):
        consistent_grid = True
    else:
        consistent_grid = allow_incomplete_grid

    return consistent_grid


def has_dimensions(message: Message) -> bool:
    """ Ensure the supplied Harmony message contains values for height and
        width of the target grid, which define the sizes of the x and y
        horizontal spatial dimensions.

    """
    return _has_all_attributes(message, ['format.height', 'format.width'])


def has_crs(message: Message) -> bool:
    """Returns true if Harmony message contains a crs."""
    target_crs = rgetattr(message, 'format.crs')
    return target_crs is not None


def has_scale_extents(message: Message) -> bool:
    """ Ensure the supplied Harmony message contains values for the minimum and
        maximum extents of the target grid in both the x and y dimensions.

    """
    scale_extent_attributes = ['format.scaleExtent.x.min',
                               'format.scaleExtent.x.max',
                               'format.scaleExtent.y.min',
                               'format.scaleExtent.y.max']

    return _has_all_attributes(message, scale_extent_attributes)


def has_scale_sizes(message: Message) -> bool:
    """ Ensure the supplied Harmony message contains values for the x and y
        horizontal scale sizes for the target grid.

    """
    scale_size_attributes = ['format.scaleSize.x', 'format.scaleSize.y']
    return _has_all_attributes(message, scale_size_attributes)


def has_valid_scale_extents(message: Message) -> bool:
    """Ensure any input scale_extents are valid."""
    if has_scale_extents(message):
        return (
            float(rgetattr(message, 'format.scaleExtent.x.min'))
            < float(rgetattr(message, 'format.scaleExtent.x.max'))
        ) and (
            float(rgetattr(message, 'format.scaleExtent.y.min'))
            < float(rgetattr(message, 'format.scaleExtent.y.max'))
        )
    return True


def _has_all_attributes(message: Message, attributes: List[str]) -> bool:
    """ Ensure that the supplied Harmony message has non-None attribute values
        for all the listed attributes.

    """
    return all(rgetattr(message, attribute_name) is not None
               for attribute_name in attributes)


def _has_consistent_dimension(message: Message, dimension_name: str) -> bool:
    """ Ensure a grid dimension has consistent values for the scale extent
        (e.g., minimum and maximum values), scale size (resolution) and
        dimension length (e.g., width or height). For the grid x dimension, the
        calculation is as follows:

        scaleSize.x = (scaleExtent.x.max - scaleExtent.x.min) / (width)

        The message scale sizes is compared to that calculated as above, to
        ensure it is within a relative tolerance (1 x 10^-3).

    """
    message_scale_size = getattr(message.format.scaleSize, dimension_name)
    scale_extent = getattr(message.format.scaleExtent, dimension_name)

    if dimension_name == 'x':
        dimension_elements = message.format.width
    else:
        dimension_elements = message.format.height

    derived_scale_size = (scale_extent.max - scale_extent.min) / dimension_elements

    return abs(message_scale_size - derived_scale_size) <= 1e-3


def rgetattr(input_object: Any, requested_attribute: str, *args) -> Any:
    """ This is a recursive version of the inbuilt `getattr` method, such that
        it can be called to retrieve nested attributes. For example:
        the Message.subset.shape within the input Harmony message.

        Note, if a default value is specified, this will be returned if any
        attribute in the specified chain is absent from the supplied object.
        Alternatively, if an absent attribute is specified and no default value
        if given in the function call, this function will return `None`.

    """
    if len(args) == 0:
        args = (None, )

    if '.' not in requested_attribute:
        result = getattr(input_object, requested_attribute, *args)
    else:
        attribute_pieces = requested_attribute.split('.')
        result = rgetattr(getattr(input_object, attribute_pieces[0], *args),
                          '.'.join(attribute_pieces[1:]), *args)

    return result



import pytest
import responses
import os

from harmony_service_lib.http import (download, is_http, localhost_url)
from unittest.mock import Mock, patch
from tests.util import config_fixture

EDL_URL = 'https://uat.urs.earthdata.nasa.gov'

@pytest.mark.parametrize('url,expected', [
    ('http://example.com', True),
    ('HTTP://YELLING.COM', True),
    ('https://nosuchagency.org', True),
    ('hTTpS://topsecret.org', True),
    ('nothttp://topsecret.org', False),
    ('httpsnope://topsecret.org', False),
    ('s3://bucketbrigade.com', False),
    ('file:///var/log/junk.txt', False),
    ('gopher://minnesota.org', False)
])
def test_is_http(url, expected):
    assert is_http(url) is expected


@pytest.mark.parametrize('url,expected', [
    ('http://example.com/ufo_sightings.nc', 'http://example.com/ufo_sightings.nc'),
    ('http://localhost:3000/jobs', 'http://mydevmachine.local.dev:3000/jobs'),
    ('s3://localghost.org/boo.gif', 's3://localghost.org/boo.gif')
])
def test_when_given_urls_localhost_url_returns_correct_url(url, expected):
    local_hostname = 'mydevmachine.local.dev'

    assert localhost_url(url, local_hostname) == expected


@pytest.fixture
def access_token(faker):
    return faker.password(length=40, special_chars=False)

@pytest.fixture
def validate_access_token_url():
    return (f'{EDL_URL}/oauth/tokens/user'
            '?token={token}&client_id={client_id}')

@pytest.fixture
def resource_server_granule_url():
    return 'https://resource.server.daac.com/foo/bar/granule.nc'


@pytest.fixture
def response_body_from_granule_url():
    return "dummy response body"


@pytest.fixture
def resource_server_redirect_url(faker):
    return ('https://n5eil11u.ecs.nsidc.org/TS1_redirect'
            f'?code={faker.password(length=64, special_chars=False)}'
            f'&state={faker.password(length=128, special_chars=False)}')

@pytest.fixture
def edl_redirect_url(faker):
    return ('https://uat.urs.earthdata.nasa.gov/oauth/authorize'
            f'?client_id={faker.password(length=22, special_chars=False)}'
            '&response_type=code'
            '&redirect_uri=https%3A%2F%2Fn5eil11u.ecs.nsidc.org%2FTS1_redirect'
            f'&state={faker.password(length=128, special_chars=False)}')


@pytest.fixture(autouse=True)
def getsize_patched(monkeypatch):
    monkeypatch.setattr(os.path, "getsize", lambda a: 0)

@responses.activate
def test_download_follows_redirect_and_uses_auth_headers(
        mocker,
        access_token,
        resource_server_granule_url,
        edl_redirect_url):

    responses.add(
        responses.GET,
        resource_server_granule_url,
        status=302,
        headers=[('Location', edl_redirect_url)]
    )
    responses.add(
        responses.GET,
        edl_redirect_url,
        status=302
    )
    destination_file = mocker.Mock()
    cfg = config_fixture()

    response = download(cfg, resource_server_granule_url, access_token, None, destination_file)

    # We should get redirected
    assert response.status_code == 302
    assert len(responses.calls) == 2

    # We should include auth headers in both requests
    request_headers = responses.calls[0].request.headers
    redirect_headers = responses.calls[1].request.headers

    assert 'Authorization' in request_headers
    assert 'Authorization' in redirect_headers
    assert 'Bearer' in request_headers['Authorization']
    assert 'Bearer' in redirect_headers['Authorization']


@responses.activate
@patch('harmony_service_lib.http.get_retry_delay', Mock(return_value = 0))
def test_download_validates_token_and_raises_exception(
        mocker,
        faker,
        validate_access_token_url):

    client_id = faker.password(length=22, special_chars=False)
    access_token = faker.password(length=42, special_chars=False)
    cfg = config_fixture()
    url = validate_access_token_url.format(
        token=access_token,
        client_id=client_id
    )

    responses.add(responses.POST, url, status=403, json={
        "error": "invalid_token",
        "error_description": "The token is either malformed or does not exist"
    })
    destination_file = mocker.Mock()

    with pytest.raises(Exception):
        download(cfg, 'https://xyzzy.com/foo/bar', access_token, None, destination_file)
        # Assert content


@responses.activate
def test_when_given_a_url_and_data_it_downloads_with_query_parameters(
        mocker,
        access_token,
        resource_server_granule_url):

    responses.add(
        responses.POST,
        resource_server_granule_url,
        status=200
    )
    destination_file = mocker.Mock()
    cfg = config_fixture()
    data = {'param': 'value'}

    response = download(cfg, resource_server_granule_url, access_token, data, destination_file)

    assert response.status_code == 200
    assert len(responses.calls) == 1
    assert responses.calls[0].request.body == 'param=value'


@responses.activate
def test_when_authn_succeeds_it_writes_to_provided_file(
        mocker,
        access_token,
        resource_server_granule_url,
        response_body_from_granule_url):

    responses.add(
        responses.GET,
        resource_server_granule_url,
        body=response_body_from_granule_url,
        status=200
    )
    destination_file = mocker.Mock()
    cfg = config_fixture()

    response = download(cfg, resource_server_granule_url, access_token, None, destination_file)

    assert response.status_code == 200
    assert len(responses.calls) == 1
    destination_file.write.assert_called()


@responses.activate
@patch('harmony_service_lib.http.get_retry_delay', Mock(return_value = 0))
def test_download_all_retries_failed(
        mocker,
        faker,
        resource_server_granule_url):

    client_id = faker.password(length=22, special_chars=False)
    access_token = faker.password(length=42, special_chars=False)
    cfg = config_fixture()

    responses.add(
        responses.GET,
        resource_server_granule_url,
        status=599
    )
    destination_file = mocker.Mock()

    with pytest.raises(Exception):
         download(cfg, resource_server_granule_url, access_token, None, destination_file)

    assert len(responses.calls) == 5

@responses.activate
def test_user_agent_is_passed_to_request_headers_when_using_edl_auth(
        mocker,
        access_token,
        resource_server_granule_url):

    responses.add(
        responses.GET,
        resource_server_granule_url,
        status=200
    )
    destination_file = mocker.Mock()
    cfg = config_fixture()

    user_agent = 'test-agent/0.0.0'
    download(cfg, resource_server_granule_url, access_token, None, destination_file, user_agent=user_agent)

    assert 'User-Agent' in responses.calls[0].request.headers
    assert user_agent in responses.calls[0].request.headers['User-Agent']

@responses.activate
def test_user_agent_is_passed_to_request_headers_when_using_edl_auth_and_post_param(
        mocker,
        access_token,
        resource_server_granule_url):

    responses.add(
        responses.POST,
        resource_server_granule_url,
        status=200
    )
    destination_file = mocker.Mock()
    cfg = config_fixture()
    data = {'param': 'value'}

    user_agent = 'test-agent/0.0.0'
    download(cfg, resource_server_granule_url, access_token, data, destination_file, user_agent=user_agent)

    assert 'User-Agent' in responses.calls[0].request.headers
    assert user_agent in responses.calls[0].request.headers['User-Agent']

RETRY_ERROR_CODES = [400, 404, 500, 502, 503]

@responses.activate(registry=responses.registries.OrderedRegistry)
@pytest.mark.parametrize('error_code', [RETRY_ERROR_CODES])
@patch('harmony_service_lib.http.get_retry_delay', Mock(return_value = 0))
def test_retries_on_temporary_errors_edl_auth(
        mocker,
        access_token,
        resource_server_granule_url,
        error_code):
    rsp1 = responses.get(resource_server_granule_url, body="Error", status=error_code)
    rsp2 = responses.get(resource_server_granule_url, body="Error", status=error_code)
    rsp3 = responses.get(resource_server_granule_url, body="OK", status=200)

    destination_file = mocker.Mock()
    cfg = config_fixture(max_download_retries=5)

    response = download(cfg, resource_server_granule_url, access_token, None, destination_file)

    assert response.status_code == 200
    assert rsp1.call_count == 1
    assert rsp2.call_count == 1
    assert rsp3.call_count == 1



from urllib.parse import urlparse
import boto3
from pystac.stac_io import StacIO, DefaultStacIO
from harmony_service_lib import util
from harmony_service_lib import aws
from os import environ

"""
Read and write to s3 when STAC links start with s3://.
https://pystac.readthedocs.io/en/latest/concepts.html#using-stac-io
"""

defaultStacIO = DefaultStacIO()


class S3StacIO(StacIO):

    def read_text(self, uri):
        """
        Reads STAC files from s3
        (or via the default method if the protocol is not s3).

        Parameters
        ----------
        uri: The STAC file uri.

        Returns
        -------
        The file contents
        """
        config = util.config(validate=environ.get('ENV') != 'test')
        service_params = aws.aws_parameters(
            config.use_localstack, config.localstack_host, config.aws_default_region)
        parsed = urlparse(uri)
        if parsed.scheme == 's3':
            bucket = parsed.netloc
            key = parsed.path[1:]
            s3 = boto3.resource('s3', **service_params)
            obj = s3.Object(bucket, key)
            return obj.get()['Body'].read().decode('utf-8')
        else:
            return defaultStacIO.read_text(uri)

    def write_text(self, uri, txt):
        """
        Writes a STAC file to the given uri.

        Parameters
        ----------
        uri: The STAC file uri.
        txt: The STAC contents.
        """
        parsed = urlparse(uri)
        if parsed.scheme == 's3':
            aws.write_s3(uri, txt)
        else:
            defaultStacIO.write_text(uri, txt)



"""
==================
example_service.py
==================

An example service adapter implementation and example CLI parser
"""

import argparse
import shutil
import os
from tempfile import mkdtemp
from pystac import Asset

import harmony_service_lib
from harmony_service_lib.util import generate_output_filename, stage, download

# IMPORTANT: The following line avoids making real calls to a non-existent
# Harmony frontend.  Service authors should not set this variable to "dev"
# or "test" when releasing the service.
os.environ['ENV'] = 'dev'


class ExampleAdapter(harmony_service_lib.BaseHarmonyAdapter):
    """
    Shows an example of what a service adapter implementation looks like
    """
    def process_item(self, item, source):
        """
        Processes a single input item.  Services that are not aggregating multiple input files
        should prefer to implement this method rather than #invoke

        This example copies its input to the output, marking "dpi" and "variables" message
        attributes as having been processed

        Parameters
        ----------
        item : pystac.Item
            the item that should be processed
        source : harmony.message.Source
            the input source defining the variables, if any, to subset from the item

        Returns
        -------
        pystac.Item
            a STAC catalog whose metadata and assets describe the service output
        """
        result = item.clone()
        result.assets = {}

        # Create a temporary dir for processing we may do
        workdir = mkdtemp()
        try:
            # Get the data file
            asset = next(v for k, v in item.assets.items() if 'data' in (v.roles or []))
            input_filename = download(asset.href, workdir, logger=self.logger, access_token=self.message.accessToken)

            # Mark any fields the service processes so later services do not repeat work
            dpi = self.message.format.process('dpi')
            # Variable subsetting
            variables = source.process('variables')

            # Do the work here!
            var_names = [v.name for v in variables]
            print('Processing item %s, DPI=%d, vars=[%s]' % (item.id, dpi, ', '.join(var_names)))
            working_filename = os.path.join(workdir, 'tmp.txt')
            shutil.copyfile(input_filename, working_filename)

            # Stage the output file with a conventional filename
            output_filename = generate_output_filename(asset.href, ext=None, variable_subset=None,
                                                       is_regridded=False, is_subsetted=False)
            url = stage(working_filename, output_filename, 'text/plain', location=self.message.stagingLocation,
                        logger=self.logger)

            # Update the STAC record
            result.assets['data'] = Asset(url, title=output_filename, media_type='text/plain', roles=['data'])
            # Other metadata updates may be appropriate, such as result.bbox and result.geometry
            # if a spatial subset was performed

            # Return the STAC record
            return result
        finally:
            # Clean up any intermediate resources
            shutil.rmtree(workdir)


def run_cli(args):
    """
    Runs the CLI.  Presently stubbed to demonstrate how a non-Harmony CLI fits in and allow
    future implementation or removal if desired.

    Parameters
    ----------
    args : Namespace
        Argument values parsed from the command line, presumably via ArgumentParser.parse_args

    Returns
    -------
    None
    """
    print("TODO: You can implement a non-Harmony CLI here.")
    print('To see the Harmony CLI, pass `--harmony-action=invoke '
          '--harmony-input="$(cat example/example_message.json)" '
          '--harmony-sources=example/source/catalog.json --harmony-output-dir=tmp/`')



def main():
    """
    Parses command line arguments and invokes the appropriate method to respond to them

    Returns
    -------
    None
    """
    parser = argparse.ArgumentParser(prog='example', description='Run an example service')

    harmony_service_lib.setup_cli(parser)

    args = parser.parse_args()

    if (harmony_service_lib.is_harmony_cli(args)):
        harmony_service_lib.run_cli(parser, args, ExampleAdapter)
    else:
        run_cli(args)


if __name__ == "__main__":
    main()



from urllib.parse import urlparse, parse_qs
from requests.auth import AuthBase
from requests import Session


def _is_presigned_url(url: str) -> bool:
    """Check if the URL is an AWS presigned URL. For AWS presigned URLs we do not
    want to pass in an authorization header."""
    query_params = parse_qs(urlparse(url).query)
    return "X-Amz-Algorithm" in query_params or "Signature" in query_params


class EarthdataSession(Session):
    """Session which ensures the Authorization header is sent to correct
    servers.

    After instantiating the EarthdataSession, set its `auth` attribute
    to a valid EarthdataAuth instance:

        session.auth = EarthdataAuth(...)

    This lifecycle method on requests.Session is called when handling
    redirect requests.
    """
    def rebuild_auth(self, prepared_request, response):
        # If not configured with an EarthdataAuth instance, defer to default behavior
        if not self.auth:
            return super().rebuild_auth(prepared_request, response)

        self.auth(prepared_request)


class EarthdataAuth(AuthBase):
    """Custom Earthdata Auth provider to add EDL Authorization headers."""

    def __init__(self, user_access_token: str):
        self.authorization_header = f"Bearer {user_access_token}"

    def __call__(self, r):
        """Add Authorization headers unless the request is to an AWS presigned URL."""
        if _is_presigned_url(r.url):
            r.headers.pop('Authorization', None)
        else:
            r.headers["Authorization"] = self.authorization_header

        return r



"""Utility functionality for generating provenance metadata."""
from __future__ import annotations

from datetime import datetime

try:
    # datetime.UTC added in Python 3.11
    from datetime import UTC
except ImportError:
    from datetime import timezone
    UTC = timezone.utc


def get_updated_history_metadata(
    service_name: str,
    service_version: str,
    existing_history: str | None = None,
) -> str:
    """Create updated history global attribute.

    This function primarily ensures the correct formatting of the history
    string, and is agnostic to the format of the input file.

    Parameters
    ----------
        service_name : str
            The name of the service processing a file.
        service_version : str
            The semantic version number of the service processing a file.
        existing_history : str, optional
            Any existing content from a "history" or "History" metadata
            attribute, to which new information should be appended.

    Returns
    -------
        str
            Updated value for history metadata attribute. If an existing
            history value was supplied, the new information is appended to that
            value, and delimited by a newline.

    """
    new_history_line = ' '.join(
        [
            datetime.now(UTC).isoformat(),
            service_name,
            service_version,
        ]
    )

    return '\n'.join(
        filter(None, [existing_history, new_history_line])
    )



"""
===========
__init__.py
===========

Convenience exports for the Harmony library
"""

# Automatically updated by `make build`
__version__ = "2.7.1"

from .adapter import BaseHarmonyAdapter
from .cli import setup_cli, is_harmony_cli, run_cli
from .message import Temporal
from pystac.stac_io import StacIO
from .s3_stac_io import S3StacIO

StacIO.set_default(S3StacIO)
